__author__ = "ajshajib"

from lenstronomy.LensModel.Profiles.gauss_decomposition import SersicEllipseGaussDec
from lenstronomy.LensModel.Profiles.gauss_decomposition import CTNFWGaussDec
from lenstronomy.LensModel.Profiles.sersic import Sersic
from lenstronomy.LightModel.Profiles.sersic import SersicElliptic

import numpy as np
import numpy.testing as npt
import pytest


class TestSersicEllipseGaussDec(object):
    """This class tests the methods for Gauss-decomposed elliptic Sersic convergence."""

    def setup_method(self):
        self.sersic_gauss = SersicEllipseGaussDec()
        self.sersic_light = SersicElliptic(sersic_major_axis=False)
        self.sersic_sphere = Sersic(sersic_major_axis=False)

    def test_function(self):
        """Test the potential function of Gauss-decomposed elliptical Sersic by
        asserting that the numerical derivative of the computed potential matches with
        the analytical derivative values.

        :return:
        :rtype:
        """
        k_eff = 1.0
        R_sersic = 1.0
        n_sersic = 1.0
        e1 = 0.2
        e2 = 0.2
        center_x = 0.0
        center_y = 0.0

        diff = 1.0e-6

        n = 5
        xs = np.linspace(0.5 * R_sersic, 2.0 * R_sersic, n)
        ys = np.linspace(0.5 * R_sersic, 2.0 * R_sersic, n)

        for x, y in zip(xs, ys):
            func = self.sersic_gauss.function(
                x,
                y,
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
                n_sersic=n_sersic,
                R_sersic=R_sersic,
                k_eff=k_eff,
            )

            func_dx = self.sersic_gauss.function(
                x + diff,
                y,
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
                n_sersic=n_sersic,
                R_sersic=R_sersic,
                k_eff=k_eff,
            )

            func_dy = self.sersic_gauss.function(
                x,
                y + diff,
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
                n_sersic=n_sersic,
                R_sersic=R_sersic,
                k_eff=k_eff,
            )

            f_x_num = (func_dx - func) / diff
            f_y_num = (func_dy - func) / diff

            f_x, f_y = self.sersic_gauss.derivatives(
                x,
                y,
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
                n_sersic=n_sersic,
                R_sersic=R_sersic,
                k_eff=k_eff,
            )

            npt.assert_almost_equal(f_x_num, f_x, decimal=4)
            npt.assert_almost_equal(f_y_num, f_y, decimal=4)

    def test_derivatives(self):
        """Test the derivative function of Gauss-decomposed elliptical Sersic by
        matching with the spherical case.

        :return:
        :rtype:
        """
        k_eff = 1.0
        R_sersic = 1.0
        n_sersic = 1.0
        e1 = 5.0e-5
        e2 = 0.0
        center_x = 0.0
        center_y = 0.0

        n = 10
        x = np.linspace(0.5 * R_sersic, 2.0 * R_sersic, n)
        y = np.linspace(0.5 * R_sersic, 2.0 * R_sersic, n)

        X, Y = np.meshgrid(x, y)

        f_x_s, f_y_s = self.sersic_sphere.derivatives(
            X,
            Y,
            center_x=center_x,
            center_y=center_y,
            n_sersic=n_sersic,
            R_sersic=R_sersic,
            k_eff=k_eff,
        )
        f_x, f_y = self.sersic_gauss.derivatives(
            X,
            Y,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
            n_sersic=n_sersic,
            R_sersic=R_sersic,
            k_eff=k_eff,
        )

        npt.assert_allclose(f_x, f_x_s, rtol=1e-3, atol=0.0)
        npt.assert_allclose(f_y, f_y_s, rtol=1e-3, atol=0.0)

        npt.assert_almost_equal(f_x, f_x_s, decimal=3)
        npt.assert_almost_equal(f_y, f_y_s, decimal=3)

    def test_hessian(self):
        """Test the Hessian function of Gauss-decomposed elliptical Sersic by matching
        with the spherical case.

        :return:
        :rtype:
        """
        k_eff = 1.0
        R_sersic = 1.0
        n_sersic = 1.0
        e1 = 5e-5
        e2 = 0.0
        center_x = 0.0
        center_y = 0.0

        n = 10
        x = np.linspace(0.5 * R_sersic, 2.0 * R_sersic, n)
        y = np.linspace(0.5 * R_sersic, 2.0 * R_sersic, n)

        X, Y = np.meshgrid(x, y)

        f_xx_s, f_xy_s, f_yx_s, f_yy_s = self.sersic_sphere.hessian(
            X,
            Y,
            center_x=center_x,
            center_y=center_y,
            n_sersic=n_sersic,
            R_sersic=R_sersic,
            k_eff=k_eff,
        )
        f_xx, f_xy, f_yx, f_yy = self.sersic_gauss.hessian(
            X,
            Y,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
            n_sersic=n_sersic,
            R_sersic=R_sersic,
            k_eff=k_eff,
        )

        npt.assert_almost_equal(f_xx_s, f_xx, decimal=3)
        npt.assert_almost_equal(f_yy_s, f_yy, decimal=3)
        npt.assert_almost_equal(f_xy_s, f_xy, decimal=3)
        npt.assert_almost_equal(f_xy_s, f_yx_s, decimal=3)

    def test_density_2d(self):
        """Test the density function of Gauss-decomposed elliptical Sersic by checking
        with the spherical case.

        :return:
        :rtype:
        """
        k_eff = 1.0
        R_sersic = 1.0
        n_sersic = 1.0
        e1 = 0.2
        e2 = 0.2
        center_x = 0.0
        center_y = 0.0

        n = 100
        x = np.logspace(-1.0, 1.0, n)
        y = np.logspace(-1.0, 1.0, n)

        X, Y = np.meshgrid(x, y)

        sersic_analytic = self.sersic_light.function(
            X,
            Y,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
            n_sersic=n_sersic,
            R_sersic=R_sersic,
            amp=k_eff,
        )

        sersic_gauss = self.sersic_gauss.density_2d(
            X,
            Y,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
            n_sersic=n_sersic,
            R_sersic=R_sersic,
            k_eff=k_eff,
        )

        print(np.abs(sersic_analytic - sersic_gauss) / np.sqrt(sersic_analytic))

        assert np.all(
            np.abs(sersic_analytic - sersic_gauss) / np.sqrt(sersic_analytic) * 100.0
            < 1.0
        )

    def test_gauss_decompose_sersic(self):
        """Test that `gauss_decompose_sersic()` decomposes the Sersic profile within 1%
        Poission noise at R_sersic.

        :return:
        :rtype:
        """
        y = np.logspace(-1.0, 1.0, 100)

        k_eff = 1.0
        R_sersic = 1.0
        n_sersic = 1.0

        amps, sigmas = self.sersic_gauss.gauss_decompose(
            n_sersic=n_sersic, R_sersic=R_sersic, k_eff=k_eff
        )

        sersic = self.sersic_gauss.get_kappa_1d(
            y, n_sersic=n_sersic, R_sersic=R_sersic, k_eff=k_eff
        )

        back_sersic = np.zeros_like(y)

        for a, s in zip(amps, sigmas):
            back_sersic += a * np.exp(-(y**2) / 2.0 / s**2)

        assert np.all(np.abs(sersic - back_sersic) / np.sqrt(sersic) * 100.0 < 1.0)


class TestCTNFWGaussDec(object):
    """This class tests the methods for Gauss-decomposed spherical cored-truncated NFW
    profile."""

    def setup_method(self):
        self.ctnfw_gauss = CTNFWGaussDec(n_sigma=15)

    def test_gauss_decompose_ctnfw(self):
        """Test the Gaussian decomposition of core-truncated NFW profile.

        :return:
        :rtype:
        """
        rho_s = 5.0
        r_s = 5.0
        r_core = 0.3
        r_trunc = 10.0
        a = 2

        r = np.logspace(-1, 1, 1000) * r_s

        beta = r_core / r_s
        tau = r_trunc / r_s

        x = r / r_s

        true_values = (
            rho_s
            * (tau * tau / (tau * tau + x * x))
            / (x**a + beta**a) ** (1.0 / a)
            / (1.0 + x) ** 2
        )

        amps, sigmas = self.ctnfw_gauss.gauss_decompose(
            r_s=r_s, r_core=r_core, r_trunc=r_trunc, rho_s=rho_s, a=a
        )

        print(len(sigmas))
        gauss_dec_values = np.zeros_like(x)
        for a, s in zip(amps, sigmas):
            gauss_dec_values += (
                a / np.sqrt(2 * np.pi) / s * np.exp(-(r**2) / 2.0 / s**2)
            )

        # test if the approximation is valid within 2%
        npt.assert_allclose(true_values, true_values, rtol=0.02)


if __name__ == "__main__":
    pytest.main()



import types


def exporter(export_self=False):
    """Export utility, modified from https://stackoverflow.com/a/41895194
    Returns export decorator, __all__ list
    """
    all_ = []
    if export_self:
        all_.append("exporter")

    def decorator(obj):
        all_.append(obj.__name__)
        return obj

    return decorator, all_


export, __all__ = exporter(export_self=True)


@export
def laconic():
    """Activate laconic / super-shortcut mode.

    Usage:
        import lenstronomy as ls
        ls.laconic()

        lens_model = ls.LensModel(...)
        source_model = ls.LightModel(...)
        image_model = ls.ImageModel(...)
    """
    short(_laconic=True)


@export
def short(_laconic=False):
    """Activate shortcut mode.

    Usage:
        import lenstronomy as ls
        ls.short()

        lens_model = ls.LensModel.lens_model.LensModel(...)
        source_model = ls.LightModel.light_model.LightModel(...)
        image_model = ls.ImSim.image_model.ImageModel(...)
    """
    import pkgutil
    import lenstronomy

    to_add = dict()
    all_modules = dict()

    for loader, module_name, is_pkg in pkgutil.walk_packages(lenstronomy.__path__):
        # This deep internal module relies heavily on 'multiprocessing',
        # which may not be installed
        if module_name == "Sampling.Pool.multiprocessing":
            continue

        # Load the module
        module = all_modules[module_name] = loader.find_module(module_name).load_module(
            module_name
        )

        if "." in module_name:
            # Submodule, e.g. Data.psf
            # Monkeypatch the parent module to make it accessible
            fragments = module_name.split(".")
            parent_module_name, child_name = ".".join(fragments[:-1]), fragments[-1]
            setattr(all_modules[parent_module_name], child_name, module)
        else:
            # Top-level module, e.g. Data: add as lenstronomy attribute
            to_add[module_name] = module

        if _laconic:
            # If the module defines an __all__, load its symbols as well.
            # (unlike import *, we do not just load everything if __all__ is missing)
            if hasattr(module, "__all__"):
                for symbol in module.__all__:
                    symbol_name = symbol
                    if isinstance(to_add.get(symbol), types.ModuleType):
                        # Key class clashing with module name
                        # (Cosmo, LensModel, LightModel, PointSource)
                        # Try to add the symbol as LensModel_:
                        symbol_name = symbol + "_"
                    if symbol_name in to_add:
                        # Name clash! Do not add the symbol
                        # or the one it clashed with.
                        del to_add[symbol_name]
                    else:
                        to_add[symbol_name] = getattr(module, symbol)

    # Make symbols accessible under lenstronomy.[x]
    for key, value in to_add.items():
        setattr(lenstronomy, key, value)
    lenstronomy.__all__ = to_add.keys()



__author__ = "Maverick-Oh"

import numpy as np
import lenstronomy.Util.util as util
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.LensModel.Profiles.epl import EPL
from lenstronomy.LensModel.Profiles.multipole import Multipole

__all__ = ["EPL_BOXYDISKY"]


class EPL_BOXYDISKY(LensProfileBase):
    """ " EPL (Elliptical Power Law) mass profile combined with Multipole with m=4, so
    that it's either purely boxy or disky with EPL's axis and Multipole's axis aligned.

    Reference to the implementation: https://ui.adsabs.harvard.edu/abs/2022A%26A...659A.127V/abstract

    Read the documentation of lenstronomy.LensModel.Profiles.epl and lenstrnomy.LensModel.Profiles.multipole for details.

    :param theta_E: Einstein radius
    :param gamma: negative power-law slope of the 3D mass distributions
    :param e1: eccentricity. For details, read lenstronomy.Util.param_util.phi_q2_ellipticity document.
    :param e2: eccentricity. For details, read lenstronomy.Util.param_util.phi_q2_ellipticity document.
    :param center_x: center of distortion
    :param center_y: center of distortion
    :param a4_a: Strength of the deviation of multipole order 4 of the elliptical isodensity contours,
    which is translated into the multipole strength from the MULTIPOLE class through a rescaling by theta_E / sqrt(q).
    Profile is disky when a4_a>0 and boxy when a4_a<0.
    """

    param_names = ["theta_E", "gamma", "e1", "e2", "center_x", "center_y", "a4_a"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
        "a4_a": -0.1,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
        "a4_a": +0.1,
    }

    def __init__(self):
        self._epl = EPL()
        self._multipole = Multipole()
        self._m = int(4)
        super(EPL_BOXYDISKY, self).__init__()

    def _param_split(self, theta_E, gamma, e1, e2, a4_a, center_x=0, center_y=0):
        """This function splits the keyword arguments for the EPL and multipole
        profiles.

        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a4_a: amplitude of the multipole mass profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: the keyword arguments for the joint profile
        """

        phi, q = param_util.ellipticity2phi_q(e1, e2)
        rescale_am = theta_E / np.sqrt(q)
        kwargs_epl = {
            "theta_E": theta_E,
            "gamma": gamma,
            "e1": e1,
            "e2": e2,
            "center_x": center_x,
            "center_y": center_y,
        }
        kwargs_multipole = {
            "m": self._m,
            "a_m": a4_a * rescale_am,
            "phi_m": phi,
            "center_x": center_x,
            "center_y": center_y,
        }
        return kwargs_epl, kwargs_multipole

    def function(self, x, y, theta_E, gamma, e1, e2, a4_a, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param a4_a: multipole strength. The profile becomes disky when a4_a>0 and boxy when a4_a<0
        :param center_x: profile center
        :param center_y: profile center
        :return: lensing potential
        """
        kwargs_epl, kwargs_multipole = self._param_split(
            theta_E, gamma, e1, e2, a4_a, center_x=center_x, center_y=center_y
        )
        f_epl = self._epl.function(x, y, **kwargs_epl)
        f_multipole = self._multipole.function(x, y, **kwargs_multipole)
        return f_epl + f_multipole

    def derivatives(self, x, y, theta_E, gamma, e1, e2, a4_a, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param a4_a: multipole strength. The profile becomes disky when a4_a>0 and boxy when a4_a<0
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        kwargs_epl, kwargs_multipole = self._param_split(
            theta_E, gamma, e1, e2, a4_a, center_x=center_x, center_y=center_y
        )
        f_x_epl, f_y_epl = self._epl.derivatives(x, y, **kwargs_epl)
        f_x_multipole, f_y_multipole = self._multipole.derivatives(
            x, y, **kwargs_multipole
        )
        f_x = f_x_epl + f_x_multipole
        f_y = f_y_epl + f_y_multipole
        return f_x, f_y

    def hessian(self, x, y, theta_E, gamma, e1, e2, a4_a, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: power law slope
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param a4_a: multipole strength. The profile becomes disky when a4_a>0 and boxy when a4_a<0
        :param center_x: profile center
        :param center_y: profile center
        :return: f_xx, f_xy, f_yx, f_yy
        """
        kwargs_epl, kwargs_multipole = self._param_split(
            theta_E, gamma, e1, e2, a4_a, center_x=center_x, center_y=center_y
        )
        f_xx_epl, f_xy_epl, f_yx_epl, f_yy_epl = self._epl.hessian(x, y, **kwargs_epl)
        (
            f_xx_multipole,
            f_xy_multipole,
            f_yx_multipole,
            f_yy_multipole,
        ) = self._multipole.hessian(x, y, **kwargs_multipole)
        f_xx = f_xx_epl + f_xx_multipole
        f_xy = f_xy_epl + f_xy_multipole
        f_yx = f_yx_epl + f_yx_multipole
        f_yy = f_yy_epl + f_yy_multipole
        return f_xx, f_xy, f_yx, f_yy



import numpy as np
from lenstronomy.LensModel.Profiles.shear import ShearReduced
from lenstronomy.LensModel.Profiles.curved_arc_sis_mst import CurvedArcSISMST
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["CurvedArcSPT"]


class CurvedArcSPT(LensProfileBase):
    """Curved arc model based on SIS+MST with an additional non-linear shear distortions
    applied on the source coordinates around the center.

    This profile is effectively a Source Position Transform of a curved arc and a shear
    distortion.
    """

    param_names = [
        "tangential_stretch",
        "radial_stretch",
        "curvature",
        "direction",
        "gamma1",
        "gamma2",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "tangential_stretch": -100,
        "radial_stretch": -5,
        "curvature": 0.000001,
        "direction": -np.pi,
        "gamma1": -0.5,
        "gamma2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "tangential_stretch": 100,
        "radial_stretch": 5,
        "curvature": 100,
        "direction": np.pi,
        "gamma1": 0.5,
        "gamma2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._curve = CurvedArcSISMST()
        self._distort = ShearReduced()
        super(CurvedArcSPT, self).__init__()

    def function(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        direction,
        gamma1,
        gamma2,
        center_x,
        center_y,
    ):
        """
        ATTENTION: there may not be a global lensing potential!

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param gamma1: non-linear reduced shear distortion in the source plane
        :param gamma2: non-linear reduced shear distortion in the source plane
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        raise NotImplemented(
            "lensing potential for regularly curved arc is not implemented"
        )

    def derivatives(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        direction,
        gamma1,
        gamma2,
        center_x,
        center_y,
    ):
        """

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param gamma1: non-linear reduced shear distortion in the source plane
        :param gamma2: non-linear reduced shear distortion in the source plane
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        # computed regular curved arc deflection
        f_x_c, f_y_c = self._curve.derivatives(
            x,
            y,
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
            center_x,
            center_y,
        )
        # map to source plane coordinate system
        beta_x, beta_y = x - f_x_c, y - f_y_c
        # distort source plane coordinate system around (center_x, center_y)
        f_x_b, f_y_b = self._distort.derivatives(
            beta_x, beta_y, gamma1, gamma2, ra_0=center_x, dec_0=center_y
        )
        beta_x_, beta_y_ = beta_x - f_x_b, beta_y - f_y_b
        # compute total deflection between initial coordinate and final source coordinate to match lens equation
        # beta = theta - alpha
        f_x, f_y = x - beta_x_, y - beta_y_
        return f_x, f_y

    def hessian(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        direction,
        gamma1,
        gamma2,
        center_x,
        center_y,
    ):
        """

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param gamma1: non-linear reduced shear distortion in the source plane
        :param gamma2: non-linear reduced shear distortion in the source plane
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        alpha_ra, alpha_dec = self.derivatives(
            x,
            y,
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
            gamma1,
            gamma2,
            center_x,
            center_y,
        )
        diff = 0.0000001
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff,
            y,
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
            gamma1,
            gamma2,
            center_x,
            center_y,
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x,
            y + diff,
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
            gamma1,
            gamma2,
            center_x,
            center_y,
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff
        return f_xx, f_xy, f_yx, f_yy



import numpy as np

__all__ = ["PointSourceParam"]

from lenstronomy.Sampling.param_group import ModelParamGroup, SingleParam, ArrayParam


class SourcePositionParam(SingleParam):
    """Source position parameter, ra_source and dec_source."""

    param_names = ["ra_source", "dec_source"]
    _kwargs_lower = {"ra_source": -100, "dec_source": -100}
    _kwargs_upper = {"ra_source": 100, "dec_source": 100}


class LensedPosition(ArrayParam):
    """Represents lensed positions, possibly many. ra_image and dec_image.

    :param num_images: integer. The number of lensed positions to model.
    """

    _kwargs_lower = {
        "ra_image": -100,
        "dec_image": -100,
    }
    _kwargs_upper = {
        "ra_image": 100,
        "dec_image": 100,
    }

    def __init__(self, num_images):
        ArrayParam.__init__(self, int(num_images) > 0)
        self.param_names = {"ra_image": int(num_images), "dec_image": int(num_images)}


class SourceAmp(SingleParam):
    """Source amplification."""

    param_names = ["source_amp"]
    _kwargs_lower = {"source_amp": 0}
    _kwargs_upper = {"source_amp": 100}


class ImageAmp(ArrayParam):
    """Observed amplification of lensed images of a point source. Can model arbitrarily
    many magnified images.

    :param num_point_sources: integer. The number of lensed images without fixed
        magnification.
    """

    _kwargs_lower = {"point_amp": 0}
    _kwargs_upper = {"point_amp": 100}

    def __init__(self, num_point_sources):
        ArrayParam.__init__(self, int(num_point_sources) > 0)
        self.param_names = {"point_amp": int(num_point_sources)}


class PointSourceParam(object):
    """Point source parameters."""

    def __init__(
        self,
        model_list,
        kwargs_fixed,
        num_point_source_list=None,
        linear_solver=True,
        fixed_magnification_list=None,
        kwargs_lower=None,
        kwargs_upper=None,
    ):
        """

        :param model_list: list of point source model names
        :param kwargs_fixed: list of keyword arguments with parameters to be held fixed
        :param num_point_source_list: list of number of point sources per point source model class
        :param linear_solver: bool, if True, does not return linear parameters for the sampler
         (will be solved linearly instead)
        :param fixed_magnification_list: list of booleans, if entry is True, keeps one overall scaling among the
         point sources in this class
        """
        self.model_list = model_list
        if num_point_source_list is None:
            num_point_source_list = [1] * len(model_list)
        self._num_point_sources_list = num_point_source_list
        if fixed_magnification_list is None:
            fixed_magnification_list = [False] * len(model_list)
        self._fixed_magnification_list = fixed_magnification_list
        self.kwargs_fixed = kwargs_fixed
        if linear_solver is True:
            self.kwargs_fixed = self.add_fix_linear(kwargs_fixed)
        self._linear_solver = linear_solver

        self.param_groups = []
        for i, model in enumerate(self.model_list):
            params = []
            num = num_point_source_list[i]
            if model in ["LENSED_POSITION", "UNLENSED"]:
                params.append(LensedPosition(num))
            elif model == "SOURCE_POSITION":
                params.append(SourcePositionParam(True))
            else:
                raise ValueError("%s not a valid point source model" % model)

            if fixed_magnification_list[i] and model in [
                "LENSED_POSITION",
                "SOURCE_POSITION",
            ]:
                params.append(SourceAmp(True))
            else:
                params.append(ImageAmp(num))

            self.param_groups.append(params)

        if kwargs_lower is None:
            kwargs_lower = []
            for model_params in self.param_groups:
                fixed_lower = {}
                for param_group in model_params:
                    fixed_lower = dict(fixed_lower, **param_group.kwargs_lower)
                kwargs_lower.append(fixed_lower)

        if kwargs_upper is None:
            kwargs_upper = []
            for model_params in self.param_groups:
                fixed_upper = {}
                for param_group in model_params:
                    fixed_upper = dict(fixed_upper, **param_group.kwargs_upper)
                kwargs_upper.append(fixed_upper)

        self.lower_limit = kwargs_lower
        self.upper_limit = kwargs_upper

    def get_params(self, args, i):
        """

        :param args: sorted list of floats corresponding to the parameters being sampled
        :param i: int, index of first entry relevant for being managed by this class
        :return: keyword argument list of point sources, index relevant for the next class
        """
        kwargs_list = []
        for k, param_group in enumerate(self.param_groups):
            kwargs, i = ModelParamGroup.compose_get_params(
                param_group, args, i, kwargs_fixed=self.kwargs_fixed[k]
            )
            kwargs_list.append(kwargs)
        return kwargs_list, i

    def set_params(self, kwargs_list):
        """

        :param kwargs_list: keyword argument list
        :return: sorted list of parameters being sampled extracted from kwargs_list
        """
        args = []
        for k, param_group in enumerate(self.param_groups):
            kwargs = kwargs_list[k]
            kwargs_fixed = self.kwargs_fixed[k]
            args.extend(
                ModelParamGroup.compose_set_params(
                    param_group, kwargs, kwargs_fixed=kwargs_fixed
                )
            )
        return args

    def num_param(self):
        """Number of parameters and their names.

        :return: int, list of parameter names
        """
        num, name_list = 0, []
        for k, param_group in enumerate(self.param_groups):
            n, names = ModelParamGroup.compose_num_params(
                param_group, kwargs_fixed=self.kwargs_fixed[k]
            )
            num += n
            name_list += names
        return num, name_list

    def add_fix_linear(self, kwargs_fixed):
        """Updates fixed keyword argument list with linear parameters.

        :param kwargs_fixed: list of keyword arguments held fixed during sampling
        :return: updated keyword argument list
        """
        for k, model in enumerate(self.model_list):
            if self._fixed_magnification_list[k] is True and model in [
                "LENSED_POSITION",
                "SOURCE_POSITION",
            ]:
                kwargs_fixed[k]["source_amp"] = 1
            else:
                kwargs_fixed[k]["point_amp"] = np.ones(self._num_point_sources_list[k])
        return kwargs_fixed

    def num_param_linear(self):
        """

        :return: number of linear parameters
        """
        num = 0
        if self._linear_solver is True:
            for k, model in enumerate(self.model_list):
                if self._fixed_magnification_list[k] is True and model in [
                    "LENSED_POSITION",
                    "SOURCE_POSITION",
                ]:
                    num += 1
                else:
                    num += self._num_point_sources_list[k]
        return num



__author__ = "sibirrer"


import numpy as np
from astropy.cosmology import default_cosmology
from lenstronomy.Util import class_creator
from lenstronomy.Util import constants as const
from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from lenstronomy.Analysis.kinematics_api import KinematicsAPI

__all__ = ["TDCosmography"]


class TDCosmography(KinematicsAPI):
    """Class equipped to perform a cosmographic analysis from a lens model with added
    measurements of time delays and kinematics.

    This class does not require any cosmological knowledge and can return angular
    diameter distance estimates self-consistently integrating the kinematics routines
    and time delay estimates in the lens modeling. This description follows Birrer et
    al. 2016, 2019.
    """

    def __init__(
        self,
        z_lens,
        z_source,
        kwargs_model,
        cosmo_fiducial=None,
        lens_model_kinematics_bool=None,
        light_model_kinematics_bool=None,
        kwargs_seeing=None,
        kwargs_aperture=None,
        anisotropy_model=None,
        **kwargs_kin_api
    ):
        """

        :param z_lens: redshift of deflector
        :param z_source: redshift of source
        :param kwargs_model: model configurations (according to FittingSequence)
        :param cosmo_fiducial: fiducial cosmology used to compute angular diameter distances where required
        :param lens_model_kinematics_bool: (optional) bool list, corresponding to lens models being included into the
         kinematics modeling
        :param light_model_kinematics_bool: (optional) bool list, corresponding to lens light models being included
         into the kinematics modeling
        :param kwargs_seeing: seeing conditions (see observation class in Galkin)
        :param kwargs_aperture: aperture keyword arguments (see aperture class in Galkin)
        :param anisotropy_model: string, anisotropy model type
        :param kwargs_kin_api: additional keyword arguments for KinematicsAPI class instance
        """

        if cosmo_fiducial is None:
            cosmo_fiducial = default_cosmology.get()
        if kwargs_seeing is None:
            kwargs_seeing = {}
        if kwargs_aperture is None:
            kwargs_aperture = {}
        self._z_lens = z_lens
        self._z_source = z_source
        self._cosmo_fiducial = cosmo_fiducial
        self._lens_cosmo = LensCosmo(
            z_lens=z_lens, z_source=z_source, cosmo=self._cosmo_fiducial
        )
        (
            self.LensModel,
            self.SourceModel,
            self.LensLightModel,
            self.PointSource,
            extinction_class,
        ) = class_creator.create_class_instances(all_models=True, **kwargs_model)
        super(TDCosmography, self).__init__(
            z_lens=z_lens,
            z_source=z_source,
            kwargs_model=kwargs_model,
            cosmo=cosmo_fiducial,
            lens_model_kinematics_bool=lens_model_kinematics_bool,
            light_model_kinematics_bool=light_model_kinematics_bool,
            kwargs_seeing=kwargs_seeing,
            kwargs_aperture=kwargs_aperture,
            anisotropy_model=anisotropy_model,
            **kwargs_kin_api
        )

    def time_delays(
        self, kwargs_lens, kwargs_ps, kappa_ext=0, original_ps_position=False
    ):
        """Predicts the time delays of the image positions given the fiducial cosmology
        relative to a straight line without lensing. Negative values correspond to
        images arriving earlier, and positive signs correspond to images arriving later.

        :param kwargs_lens: lens model parameters
        :param kwargs_ps: point source parameters
        :param kappa_ext: external convergence (optional)
        :param original_ps_position: boolean (only applies when first point source model
            is of type 'LENSED_POSITION'), uses the image positions in the model
            parameters and does not re-compute images (which might be differently
            ordered) in case of the lens equation solver
        :return: time delays at image positions for the fixed cosmology in units of days
        """
        fermat_pot = self.fermat_potential(
            kwargs_lens, kwargs_ps, original_ps_position=original_ps_position
        )
        time_delay = self._lens_cosmo.time_delay_units(fermat_pot, kappa_ext)
        return time_delay

    def fermat_potential(self, kwargs_lens, kwargs_ps, original_ps_position=False):
        """Fermat potential (negative sign means earlier arrival time)

        :param kwargs_lens: lens model keyword argument list
        :param kwargs_ps: point source keyword argument list
        :param original_ps_position: boolean (only applies when first point source model
            is of type 'LENSED_POSITION'), uses the image positions in the model
            parameters and does not re-compute images (which might be differently
            ordered) in case of the lens equation solver
        :return: Fermat potential of all the image positions in the first point source
            list entry
        """
        ra_pos, dec_pos = self.PointSource.image_position(
            kwargs_ps, kwargs_lens, original_position=original_ps_position
        )
        ra_pos = ra_pos[0]
        dec_pos = dec_pos[0]
        ra_source, dec_source = self.LensModel.ray_shooting(
            ra_pos, dec_pos, kwargs_lens
        )
        sigma_source = np.sqrt(np.var(ra_source) + np.var(dec_source))
        if sigma_source > 0.001:
            Warning(
                "Source position computed from the different image positions do not trace back to the same position! "
                "The error is %s mas and may be larger than what is required for an accurate relative time delay estimate!"
                "See e.g. Birrer & Treu 2019." % sigma_source * 1000
            )
        ra_source = np.mean(ra_source)
        dec_source = np.mean(dec_source)
        fermat_pot = self.LensModel.fermat_potential(
            ra_pos, dec_pos, kwargs_lens, ra_source, dec_source
        )
        return fermat_pot

    def velocity_dispersion_dimension_less(
        self,
        kwargs_lens,
        kwargs_lens_light,
        kwargs_anisotropy,
        r_eff=None,
        theta_E=None,
        gamma=None,
    ):
        """Sigma**2 = Dd/Dds * c**2 * J(kwargs_lens, kwargs_light, anisotropy) (Equation
        4.11 in Birrer et al. 2016 or Equation 6 in Birrer et al. 2019) J() is a
        dimensionless and cosmological independent quantity only depending on angular
        units. This function returns J given the lens and light parameters and the
        anisotropy choice without an external mass sheet correction.

        :param kwargs_lens: lens model keyword arguments
        :param kwargs_lens_light: lens light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :param r_eff: projected half-light radius of the stellar light associated with
            the deflector galaxy, optional, if set to None will be computed in this
            function with default settings that may not be accurate.
        :param theta_E: pre-computed Einstein radius (optional)
        :param gamma: pre-computed power-law slope of mass profile
        :return: dimensionless velocity dispersion (see e.g. Birrer et al. 2016, 2019)
        """
        sigma_v = self.velocity_dispersion(
            kwargs_lens=kwargs_lens,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_anisotropy=kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=gamma,
        )
        sigma_v *= 1000  # convert from [km/s] to  [m/s]
        J = sigma_v**2 * self._lens_cosmo.dds / self._lens_cosmo.ds / const.c**2
        return J

    def velocity_dispersion_map_dimension_less(
        self,
        kwargs_lens,
        kwargs_lens_light,
        kwargs_anisotropy,
        r_eff=None,
        theta_E=None,
        gamma=None,
    ):
        """Sigma**2 = Dd/Dds * c**2 * J(kwargs_lens, kwargs_light, anisotropy) (Equation
        4.11 in Birrer et al. 2016 or Equation 6 in Birrer et al. 2019) J() is a
        dimensionless and cosmological independent quantity only depending on angular
        units. This function returns J given the lens and light parameters and the
        anisotropy choice without an external mass sheet correction. This routine
        computes the IFU map of the kinematic quantities.

        :param kwargs_lens: lens model keyword arguments
        :param kwargs_lens_light: lens light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :param r_eff: projected half-light radius of the stellar light associated with
            the deflector galaxy, optional, if set to None will be computed in this
            function with default settings that may not be accurate.
        :return: dimensionless velocity dispersion (see e.g. Birrer et al. 2016, 2019)
        """
        sigma_v_map = self.velocity_dispersion_map(
            kwargs_lens=kwargs_lens,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_anisotropy=kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=gamma,
        )
        sigma_v_map *= 1000  # convert from [km/s] to  [m/s]
        J_map = (
            sigma_v_map**2 * self._lens_cosmo.dds / self._lens_cosmo.ds / const.c**2
        )
        return J_map

    @staticmethod
    def ddt_from_time_delay(
        d_fermat_model, dt_measured, kappa_s=0, kappa_ds=0, kappa_d=0
    ):
        """Time-delay distance in units of Mpc from the modeled Fermat potential and
        measured time delay from an image pair.

        :param d_fermat_model: relative Fermat potential between two images from the
            same source in units arcsec^2
        :param dt_measured: measured time delay between the same image pair in units of
            days
        :param kappa_s: external convergence from observer to source
        :param kappa_ds: external convergence from lens to source
        :param kappa_d: external convergence form observer to lens
        :return: D_dt, time-delay distance
        """
        D_dt_model = (
            dt_measured
            * const.day_s
            * const.c
            / const.Mpc
            / d_fermat_model
            / const.arcsec**2
        )
        D_dt = D_dt_model * (1 - kappa_ds) / (1 - kappa_s) / (1 - kappa_d)
        return D_dt

    @staticmethod
    def ds_dds_from_kinematics(sigma_v, J, kappa_s=0, kappa_ds=0):
        """Computes the estimate of the ratio of angular diameter distances Ds/Dds from
        the kinematic estimate of the lens and the measured dispersion.

        :param sigma_v: velocity dispersion [km/s]
        :param J: dimensionless kinematic constraint (see Birrer et al. 2016, 2019)
        :return: Ds/Dds
        """
        ds_dds_model = (sigma_v * 1000) ** 2 / const.c**2 / J
        ds_dds = ds_dds_model * (1 - kappa_ds) / (1 - kappa_s)
        return ds_dds

    def ddt_dd_from_time_delay_and_kinematics(
        self,
        d_fermat_model,
        dt_measured,
        sigma_v_measured,
        J,
        kappa_s=0,
        kappa_ds=0,
        kappa_d=0,
    ):
        """

        :param d_fermat_model: relative Fermat potential in units arcsec^2
        :param dt_measured: measured relative time delay [days]
        :param sigma_v_measured: 1-sigma Gaussian uncertainty in the measured velocity dispersion
        :param J: modeled dimensionless kinematic estimate
        :param kappa_s: LOS convergence from observer to source
        :param kappa_ds: LOS convergence from deflector to source
        :param kappa_d: LOS convergence from observer to deflector
        :return: D_dt, D_d
        """
        ddt = self.ddt_from_time_delay(
            d_fermat_model,
            dt_measured,
            kappa_s=kappa_s,
            kappa_ds=kappa_ds,
            kappa_d=kappa_d,
        )
        ds_dds = self.ds_dds_from_kinematics(
            sigma_v_measured, J, kappa_s=kappa_s, kappa_ds=kappa_ds
        )
        dd = ddt / ds_dds / (1 + self._z_lens)
        return ddt, dd



__all__ = ["LensProfileBase"]


class LensProfileBase(object):
    """This class acts as the base class of all lens model functions and indicates raise
    statements and default outputs if these functions are not defined in the specific
    lens model class."""

    def __init__(self, *args, **kwargs):
        self._static = False

    def function(self, *args, **kwargs):
        """Lensing potential (only needed for specific calculations, such as time
        delays)

        :param kwargs: keywords of the profile
        :return: raise as definition is not defined
        """
        raise ValueError(
            "function definition is not defined in the profile you want to execute."
        )

    def derivatives(self, *args, **kwargs):
        """Deflection angles.

        :param kwargs: keywords of the profile
        :return: raise as definition is not defined
        """
        raise ValueError(
            "derivatives definition is not defined in the profile you want to execute."
        )

    def hessian(self, *args, **kwargs):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2.

        :param kwargs: keywords of the profile
        :return: raise as definition is not defined
        """
        raise ValueError(
            "hessian definition is not defined in the profile you want to execute."
        )

    def density_lens(self, *args, **kwargs):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity. (optional definition)

        .. math::
            \\kappa(x, y) = \\int_{-\\infty}^{\\infty} \\rho(x, y, z) dz

        :param kwargs: keywords of the profile
        :return: raise as definition is not defined
        """
        raise ValueError(
            "density_lens definition is not defined in the profile you want to execute."
        )

    def mass_3d_lens(self, *args, **kwargs):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units The input parameter are identical as for the derivatives
        definition. (optional definition)

        :param kwargs: keywords of the profile
        :return: raise as definition is not defined
        """
        raise ValueError(
            "mass_3d_lens definition is not defined in the profile you want to execute."
        )

    def mass_2d_lens(self, *args, **kwargs):
        """Two-dimensional enclosed mass at radius r (optional definition)

        .. math::
            M_{2d}(R) = \\int_{0}^{R} \\rho_{2d}(r) 2\\pi r dr

        with :math:`\\rho_{2d}(r)` is the density_2d_lens() definition

        The mass definition is such that:

        .. math::
            \\alpha = mass_2d / r / \\pi

        with alpha is the deflection angle

        :param kwargs: keywords of the profile
        :return: raise as definition is not defined
        """
        raise ValueError(
            "mass_2d_lens definition is not defined in the profiel you want to execute."
        )

    def set_static(self, **kwargs):
        """Pre-computes certain computations that do only relate to the lens model
        parameters and not to the specific position where to evaluate the lens model.

        :param kwargs: lens model parameters
        :return: no return, for certain lens model some private self variables are
            initiated
        """
        pass

    def set_dynamic(self):
        """

        :return: no return, deletes pre-computed variables for certain lens models
        """
        pass



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.arc_perturbations import ArcPerturbations
from lenstronomy.Util import util
import numpy as np
import numpy.testing as npt
import pytest


class TestArcPerturbations(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.model = ArcPerturbations()
        self.kwargs_lens = {
            "coeff": 2,
            "d_r": 0.2,
            "d_phi": np.pi,
            "center_x": 0,
            "center_y": 0,
        }

    def test_function(self):
        x, y = util.make_grid(numPix=100, deltapix=0.1)
        values = self.model.function(x, y, **self.kwargs_lens)
        # import matplotlib.pyplot as plt
        # plt.matshow(util.array2image(values))
        # plt.show()
        npt.assert_almost_equal(values[0], 0, decimal=5)

    def test_derivatives(self):
        x, y = util.make_grid(numPix=100, deltapix=0.1)
        dx = 0.000001
        values = self.model.function(x, y, **self.kwargs_lens)
        alpha_x, alpha_y = self.model.derivatives(x, y, **self.kwargs_lens)
        values_dx = self.model.function(x + dx, y, **self.kwargs_lens)
        f_x_num = (values_dx - values) / dx
        npt.assert_almost_equal(f_x_num, alpha_x, decimal=3)
        dy = 0.000001
        values_dy = self.model.function(x, y + dy, **self.kwargs_lens)
        f_y_num = (values_dy - values) / dy
        npt.assert_almost_equal(f_y_num, alpha_y, decimal=3)

    def test_hessian(self):
        x, y = util.make_grid(numPix=100, deltapix=0.1)
        delta = 0.0000001
        f_xx, f_xy, f_yx, f_yy = self.model.hessian(x, y, **self.kwargs_lens)
        f_x, f_y = self.model.derivatives(x, y, **self.kwargs_lens)
        f_x_dx, f_y_dx = self.model.derivatives(x + delta, y, **self.kwargs_lens)
        f_x_dy, f_y_dy = self.model.derivatives(x, y + delta, **self.kwargs_lens)
        f_xx_num = (f_x_dx - f_x) / delta

        npt.assert_almost_equal(f_xx_num, f_xx, decimal=3)
        f_yy_num = (f_y_dy - f_y) / delta
        npt.assert_almost_equal(f_yy_num, f_yy, decimal=3)
        f_xy_num = (f_x_dy - f_x) / delta
        npt.assert_almost_equal(f_xy_num, f_xy, decimal=2)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import time
import copy
from lenstronomy.Sampling.Pool.pool import choose_pool
from lenstronomy.ImSim.MultiBand.single_band_multi_model import SingleBandMultiModel
from lenstronomy.Sampling.Samplers.pso import ParticleSwarmOptimizer

__all__ = ["AlignmentFitting", "AlignmentLikelihood"]


class AlignmentFitting(object):
    """Class which executes the different sampling  methods."""

    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        kwargs_params,
        band_index=0,
        likelihood_mask_list=None,
        align_offset=True,
        align_rotation=False,
    ):
        """Initialise the classes of the chain and for parameter options.

        :param align_offset: aligns shift in Ra and Dec
        :type align_offset: boolean
        :param align_rotation: aligns coordinate rotation
        :type align_rotation: boolean
        """
        self.chain = AlignmentLikelihood(
            multi_band_list,
            kwargs_model,
            kwargs_params,
            band_index,
            likelihood_mask_list,
            align_rotation=align_rotation,
            align_offset=align_offset,
        )

    def pso(
        self,
        n_particles=10,
        n_iterations=10,
        delta_shift=0.2,
        delta_rot=0.1,
        threadCount=1,
        mpi=False,
        print_key="default",
    ):
        """Returns the best fit for the lens model on catalogue basis with particle
        swarm optimizer.

        :param n_particles:
        :param n_iterations:
        :param delta_shift: astrometric shift tolerance
        :param delta_rot: rotation angle tolerance [in radian]
        :param threadCount:
        :param mpi:
        :param print_key:
        :return:
        """
        init_pos = self.chain.get_args(self.chain.kwargs_data_init)
        lower_limit, upper_limit = self.chain.lower_upper_limit(delta_shift, delta_rot)
        pool = choose_pool(mpi=mpi, processes=threadCount, use_dill=True)

        pso = ParticleSwarmOptimizer(
            self.chain, lower_limit, upper_limit, n_particles, pool=pool
        )
        if init_pos is not None:
            pso.set_global_best(
                init_pos, [0] * len(init_pos), self.chain.likelihood(init_pos)
            )

        if pool.is_master():
            print("Computing the %s ..." % print_key)

        time_start = time.time()

        result, [chi2_list, pos_list, vel_list] = pso.optimize(n_iterations)

        kwargs_data = self.chain.update_data(result)

        if pool.is_master():
            time_end = time.time()
            print("Shifts found: ", result)
            print(time_end - time_start, "time used for ", print_key)
        return kwargs_data, [chi2_list, pos_list, vel_list]


class AlignmentLikelihood(object):
    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        kwargs_params,
        band_index=0,
        likelihood_mask_list=None,
        align_offset=True,
        align_rotation=False,
    ):
        """Initializes all the classes needed for the chain.

        :param align_offset: aligns shift in Ra and Dec
        :type align_offset: boolean
        :param align_rotation: aligns coordinate rotation
        :type align_rotation: boolean
        """
        # print('initialized on cpu', threading.current_thread())
        self._align_offset = align_offset
        self._align_rotation = align_rotation
        self._multi_band_list = multi_band_list
        self.kwargs_data_init = multi_band_list[band_index][0]
        self._kwargs_data_shifted = copy.deepcopy(self.kwargs_data_init)

        self._kwargs_model = kwargs_model
        self._source_marg = False
        self._band_index = band_index
        self._likelihood_mask_list = likelihood_mask_list
        self._kwargs_params = kwargs_params

    def _likelihood(self, args):
        """Routine to compute X2 given variable parameters for a MCMC/PSO chainF."""
        # generate image and computes likelihood
        multi_band_list = self.update_multi_band(args)
        image_model = SingleBandMultiModel(
            multi_band_list,
            self._kwargs_model,
            likelihood_mask_list=self._likelihood_mask_list,
            band_index=self._band_index,
        )
        log_likelihood = image_model.likelihood_data_given_model(
            source_marg=self._source_marg, **self._kwargs_params
        )
        return log_likelihood

    def __call__(self, a):
        return self._likelihood(a)

    def likelihood(self, a):
        return self._likelihood(a)

    def setup(self):
        pass

    def update_multi_band(self, args):
        """
        :param args: list of parameters
        :return: updated multi_band_list
        """
        kwargs_data = self.update_data(args)
        multi_band_list = self._multi_band_list
        multi_band_list[self._band_index][0] = kwargs_data
        return multi_band_list

    def update_data(self, args):
        """
        :param args:
        :return:
        """
        k = 0
        kwargs_data = self._kwargs_data_shifted
        if self._align_offset:
            kwargs_data["ra_shift"] = args[k]
            kwargs_data["dec_shift"] = args[k + 1]
            k += 2
        if self._align_rotation:
            kwargs_data["phi_rot"] = args[k]
            k += 1
        return kwargs_data

    def get_args(self, kwargs_data):
        """
        :param kwargs_data: keyword arguments for ImageData()
        :return: arguments being sampled
        """
        args = []
        if self._align_offset:
            args.append(kwargs_data.get("ra_shift", 0))
            args.append(kwargs_data.get("dec_shift", 0))
        if self._align_rotation:
            args.append(kwargs_data.get("phi_rot", 0))
        return args

    @property
    def num_param(self):
        n = 0
        if self._align_offset:
            n += 2
        if self._align_rotation:
            n += 1
        return n

    def lower_upper_limit(self, delta_shift, delta_rot):
        """

        :param delta_shift: astrometric shift tolerance
        :param delta_rot: rotation angle tolerance [in radian]
        :return: lower_limit, upper_limit
        """
        lower_limit, upper_limit = [], []
        if self._align_offset:
            lower_limit.append(-delta_shift)
            lower_limit.append(-delta_shift)
            upper_limit.append(delta_shift)
            upper_limit.append(delta_shift)
        if self._align_rotation:
            lower_limit.append(-delta_rot)
            upper_limit.append(delta_rot)
        return lower_limit, upper_limit



from lenstronomy.Util.util import fwhm2sigma
from lenstronomy.LightModel.light_model import LightModel
import numpy as np


def auto_raytracing_grid_size(source_fwhm_parcsec, grid_size_scale=0.005, power=1.0):
    """This function returns the size of a ray tracing grid in units of arcsec
    appropriate for magnification computations with finite-size background sources. This
    fit is calibrated for source sizes (interpreted as the FWHM of a Gaussian) in the
    range 0.1 -100 pc.

    :param source_fwhm_parcsec: the full width at half max of a Gaussian background
        source
    :return: an appropriate grid size for finite-size background magnification
        computation
    """

    grid_radius_arcsec = grid_size_scale * source_fwhm_parcsec**power
    return grid_radius_arcsec


def auto_raytracing_grid_resolution(
    source_fwhm_parcsec, grid_resolution_scale=0.0002, ref=10.0, power=1.0
):
    """This function returns a resolution factor in units arcsec/pixel appropriate for
    magnification computations with finite-size background sources. This fit is
    calibrated for source sizes (interpreted as the FWHM of a Gaussian) in the range
    0.1-100 pc.

    :param source_fwhm_parcsec: the full width at half max of a Gaussian background
        source
    :return: an appropriate grid resolution for finite-size background magnification
        computation
    """

    grid_resolution = grid_resolution_scale * (source_fwhm_parcsec / ref) ** power
    return grid_resolution


def setup_mag_finite(
    cosmo,
    lens_model,
    grid_radius_arcsec,
    grid_resolution,
    source_fwhm_parsec,
    source_light_model,
    z_source,
    source_x,
    source_y,
    dx,
    dy,
    amp_scale,
    size_scale,
):
    """Sets up the ray tracing grid and source light model for
    magnification_finite_adaptive and plot_quasar_images routines :param cosmo:
    (optional) an instance of astropy.cosmology; if not specified, a default cosmology
    will be used :param lens_model: an instance of LensModel :param grid_radius_arcsec:
    (optional) the size of the ray tracing region in arcsec; if not specified, an
    appropriate value will be estimated from the source size :param grid_resolution: the
    grid resolution in units arcsec/pixel; if not specified, an appropriate value will
    be estimated from the source size :param source_fwhm_parsec: the size of the
    background source [units parsec] :param source_light_model: the model for background
    source light; currently implemented are 'SINGLE_GAUSSIAN' and 'DOUBLE_GAUSSIAN'.

    :param z_source: source redshift
    :param source_x: source x position [arcsec]
    :param source_y: source y position [arcsec]
    :param dx: used with source model 'DOUBLE_GAUSSIAN', the offset of the second source
        light profile from the first [arcsec]
    :param dy: used with source model 'DOUBLE_GAUSSIAN', the offset of the second source
        light profile from the first [arcsec]
    :param amp_scale: used with source model 'DOUBLE_GAUSSIAN', the peak brightness of
        the second source light profile relative to the first
    :param size_scale: used with source model 'DOUBLE_GAUSSIAN', the size of the second
        source light profile relative to the first
    :return: x coordinate grid, y coordinate grid, source light model, and keywords for
        the source light model
    """
    if cosmo is None:
        cosmo = lens_model.cosmo

    if grid_radius_arcsec is None:
        grid_radius_arcsec = auto_raytracing_grid_size(source_fwhm_parsec)
    if grid_resolution is None:
        grid_resolution = auto_raytracing_grid_resolution(source_fwhm_parsec)

    pc_per_arcsec = 1000 / cosmo.arcsec_per_kpc_proper(z_source).value
    source_fwhm_arcsec = source_fwhm_parsec / pc_per_arcsec
    source_sigma_arcsec = fwhm2sigma(source_fwhm_arcsec)

    if source_light_model == "SINGLE_GAUSSIAN":
        kwargs_source = [
            {
                "amp": 1.0,
                "center_x": source_x,
                "center_y": source_y,
                "sigma": source_sigma_arcsec,
            }
        ]
        source_model = LightModel(["GAUSSIAN"])
    elif source_light_model == "DOUBLE_GAUSSIAN":
        amp_1 = 1.0
        kwargs_source_1 = [
            {
                "amp": amp_1,
                "center_x": source_x,
                "center_y": source_y,
                "sigma": source_sigma_arcsec,
            }
        ]
        # c = amp / (2 * np.pi * sigma**2)
        amp_2 = amp_1 * amp_scale * size_scale**2
        kwargs_source_2 = [
            {
                "amp": amp_2,
                "center_x": source_x + dx,
                "center_y": source_y + dy,
                "sigma": source_sigma_arcsec * size_scale,
            }
        ]
        kwargs_source = kwargs_source_1 + kwargs_source_2
        source_model = LightModel(["GAUSSIAN"] * 2)
    else:
        raise Exception(
            "source light model must be specified, currently implemented models are  SINGLE_GAUSSIAN "
            "and DOUBLE_GAUSSIAN"
        )

    npix = int(2 * grid_radius_arcsec / grid_resolution)
    _grid_x = np.linspace(-grid_radius_arcsec, grid_radius_arcsec, npix)
    _grid_y = np.linspace(-grid_radius_arcsec, grid_radius_arcsec, npix)
    grid_x_0, grid_y_0 = np.meshgrid(_grid_x, _grid_y)

    return (
        grid_x_0,
        grid_y_0,
        source_model,
        kwargs_source,
        grid_resolution,
        grid_radius_arcsec,
    )



__author__ = "gipagano"

import numpy as np
import lenstronomy.Util.util as util
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["NIE_POTENTIAL", "NIEPotentialMajorAxis"]


class NIE_POTENTIAL(LensProfileBase):
    """This class implements the elliptical potential of Eq.

    (67) of `LECTURES ON GRAVITATIONAL LENSING <https://arxiv.org/pdf/astro-ph/9606001.pdf>`_
    and Eq. (1) of `Blandford & Kochanek 1987 <http://adsabs.harvard.edu/full/1987ApJ...321..658B>`_,
    mapped to Eq. (8) of `Barnaka1998 <https://iopscience.iop.org/article/10.1086/305950/fulltext/37798.text.html>`_
    to find the ellipticity bounds
    """

    param_names = ["center_x", "center_y", "theta_E", "theta_c", "e1", "e2"]
    lower_limit_default = {
        "center_x": -100,
        "center_y": -100,
        "theta_E": 0,
        "theta_c": 0,
        "e1": 0,
        "e2": 0,
    }
    upper_limit_default = {
        "center_x": 100,
        "center_y": 100,
        "theta_E": 10,
        "theta_c": 10,
        "e1": 0.2,
        "e2": 0.2,
    }

    def __init__(self):
        self.nie_potential_major_axis = NIEPotentialMajorAxis()
        super(NIE_POTENTIAL, self).__init__()

    def param_conv(self, theta_E, theta_c, e1, e2):
        if self._static is True:
            return (
                self._thetaE_transf_static,
                self._thetac_static,
                self._eps_static,
                self._phi_G_static,
            )
        return self._param_conv(theta_E, theta_c, e1, e2)

    def _param_conv(self, theta_E, theta_c, e1, e2):
        """Convert the spherical averaged Einstein radius to an elliptical (major axis)
        Einstein radius and the individual eccentricities to the modulus of the
        eccentricity.

        :param theta_E: Einstein radius
        :param theta_c: core radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :return: transformed Einstein radius, core radius, ellipticity modulus,
            orientation angle phi_G
        """

        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        theta_E_conv = self._theta_q_convert(theta_E, q)
        theta_c_conv = self._theta_q_convert(theta_c, q)
        return theta_E_conv, theta_c_conv, eps, phi_G

    def set_static(self, theta_E, theta_c, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param theta_c: core radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: profile center
        :param center_y: profile center
        :return: self variables set
        """
        self._static = True
        (
            self._thetaE_transf_static,
            self._thetac_static,
            self._eps_static,
            self._phi_G_static,
        ) = self._param_conv(theta_E, theta_c, e1, e2)

    def set_dynamic(self):
        """

        :return:
        """
        self._static = False
        if hasattr(self, "_thetaE_transf_static"):
            del self._thetaE_transf_static
        if hasattr(self, "_thetac_static"):
            del self._thetac_static
        if hasattr(self, "_eps_static"):
            del self._eps_static
        if hasattr(self, "_phi_G_static"):
            del self._phi_G_static

    def function(self, x, y, theta_E, theta_c, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param theta_E: Einstein radius (in angles)
        :param theta_c: core radius  (in angles)
        :param e1: eccentricity component, x direction(dimensionless)
        :param e2: eccentricity component, y direction (dimensionless)
        :return: lensing potential
        """
        theta_E_conv, theta_c_conv, eps, phi_G = self.param_conv(
            theta_E, theta_c, e1, e2
        )

        # shift
        x_ = x - center_x
        y_ = y - center_y

        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)

        # evaluate
        f_ = self.nie_potential_major_axis.function(
            x__, y__, theta_E_conv, theta_c_conv, eps
        )

        # rotate back
        return f_

    def derivatives(self, x, y, theta_E, theta_c, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param theta_E: Einstein radius (in angles)
        :param theta_c: core radius  (in angles)
        :param e1: eccentricity component, x direction(dimensionless)
        :param e2: eccentricity component, y direction (dimensionless)
        :return: deflection angle (in angles)
        """
        theta_E_conv, theta_c_conv, eps, phi_G = self.param_conv(
            theta_E, theta_c, e1, e2
        )

        # shift
        x_ = x - center_x
        y_ = y - center_y

        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)

        # evaluate
        f__x, f__y = self.nie_potential_major_axis.derivatives(
            x__, y__, theta_E_conv, theta_c_conv, eps
        )

        # rotate back
        f_x, f_y = util.rotate(f__x, f__y, -phi_G)
        return f_x, f_y

    def hessian(self, x, y, theta_E, theta_c, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param theta_E: Einstein radius (in angles)
        :param theta_c: core radius  (in angles)
        :param e1: eccentricity component, x direction(dimensionless)
        :param e2: eccentricity component, y direction (dimensionless)
        :return: hessian matrix (in angles)
        """
        theta_E_conv, theta_c_conv, eps, phi_G = self.param_conv(
            theta_E, theta_c, e1, e2
        )

        # shift
        x_ = x - center_x
        y_ = y - center_y

        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)

        # evaluate
        f__xx, f__xy, _, f__yy = self.nie_potential_major_axis.hessian(
            x__, y__, theta_E_conv, theta_c_conv, eps
        )

        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_G) * gamma1__ - np.sin(2 * phi_G) * gamma2__
        gamma2 = +np.sin(2 * phi_G) * gamma1__ + np.cos(2 * phi_G) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def _theta_q_convert(self, theta_E, q):
        """Converts a spherical averaged Einstein radius/core radius to an elliptical
        (major axis) Einstein radius. This then follows the convention of the SPEMD
        profile in lenstronomy. (theta_E / theta_E_gravlens) = sqrt[ (1+q^2) / (2 q) ]

        :param theta_E: Einstein radius in lenstronomy conventions
        :param q: axis ratio minor/major
        :return: theta_E in convention of kappa=  b *(q2(s2 + x2) + y2􏰉)−1/2
        """
        theta_E_new = theta_E / (np.sqrt((1.0 + q**2) / (2.0 * q)))  # / (1+(1-q)/2.)
        return theta_E_new


class NIEPotentialMajorAxis(LensProfileBase):
    """This class implements the elliptical potential of Eq.

    (67) of `LECTURES ON GRAVITATIONAL LENSING <https://arxiv.org/pdf/astro-ph/9606001.pdf>`_
    and Eq. (1) of `Blandford & Kochanek 1987 <http://adsabs.harvard.edu/full/1987ApJ...321..658B>`_,
    mapped to Eq. (8) of `Barnaka1998 <https://iopscience.iop.org/article/10.1086/305950/fulltext/37798.text.html>`_
    to find the ellipticity bounds
    """

    param_names = ["theta_E", "theta_c", "eps", "center_x", "center_y"]

    def __init__(self, diff=0.0000000001):
        self._diff = diff
        super(NIEPotentialMajorAxis, self).__init__()

    def function(self, x, y, theta_E, theta_c, eps):
        f_ = theta_E * np.sqrt(theta_c**2 + (1 - eps) * x**2 + (1 + eps) * y**2)
        return f_

    def derivatives(self, x, y, theta_E, theta_c, eps):
        """Returns df/dx and df/dy of the function."""
        factor = np.sqrt(theta_c**2 + (1 - eps) * x**2 + (1 + eps) * y**2)
        f_x = (theta_E / factor) * (1 - eps) * x
        f_y = (theta_E / factor) * (1 + eps) * y
        return f_x, f_y

    def hessian(self, x, y, theta_E, theta_c, eps):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        factor = np.sqrt(theta_c**2 + (1 - eps) * x**2 + (1 + eps) * y**2)
        f_xx = (1 - eps) * (theta_E / factor) - (theta_E / factor**3) * (
            1 - eps
        ) ** 2 * x**2
        f_yy = (1 + eps) * (theta_E / factor) - (theta_E / factor**3) * (
            1 + eps
        ) ** 2 * y**2
        f_xy = -(theta_E / factor**3) * (1 - eps**2) * x * y
        return f_xx, f_xy, f_xy, f_yy



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.LensModel.Profiles.curved_arc_spp import CurvedArcSPP
from lenstronomy.LensModel.Profiles.spp import SPP
from lenstronomy.LensModel.lens_model import LensModel


class TestCurvedArc(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.model = CurvedArcSPP()
        self.spp = SPP()

    def test_spp2stretch(self):
        center_x, center_y = 1, 1
        theta_E = 1
        gamma = 1.9
        center_x_spp, center_y_spp = 0.0, 0

        (
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
        ) = self.model.spp2stretch(
            theta_E, gamma, center_x_spp, center_y_spp, center_x, center_y
        )
        (
            theta_E_new,
            gamma_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(gamma_new, gamma, decimal=8)

        center_x, center_y = -1, 1
        (
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
        ) = self.model.spp2stretch(
            theta_E, gamma, center_x_spp, center_y_spp, center_x, center_y
        )
        (
            theta_E_new,
            gamma_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(gamma_new, gamma, decimal=8)

        center_x, center_y = 0, 0.5
        (
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
        ) = self.model.spp2stretch(
            theta_E, gamma, center_x_spp, center_y_spp, center_x, center_y
        )
        (
            theta_E_new,
            gamma_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(gamma_new, gamma, decimal=8)

        center_x, center_y = 0, -1.5
        (
            tangential_stretch,
            radial_stretch,
            r_curvature,
            direction,
        ) = self.model.spp2stretch(
            theta_E, gamma, center_x_spp, center_y_spp, center_x, center_y
        )
        print(tangential_stretch, radial_stretch, r_curvature, direction)
        (
            theta_E_new,
            gamma_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2spp(
            tangential_stretch,
            radial_stretch,
            r_curvature,
            direction,
            center_x,
            center_y,
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(gamma_new, gamma, decimal=8)

    def test_function(self):
        center_x, center_y = 0.0, 0.0
        x, y = 1, 1
        output = self.model.function(
            x,
            y,
            tangential_stretch=2,
            radial_stretch=1,
            curvature=1.0 / 2,
            direction=0,
            center_x=0,
            center_y=0,
        )
        theta_E, gamma, center_x_spp, center_y_spp = self.model.stretch2spp(
            tangential_stretch=2,
            radial_stretch=1,
            curvature=1.0 / 2,
            direction=0,
            center_x=0,
            center_y=0,
        )
        out_spp = self.spp.function(1, 1, theta_E, gamma, center_x_spp, center_y_spp)
        alpha_x, alpha_y = self.spp.derivatives(
            center_x, center_y, theta_E, gamma, center_x_spp, center_y_spp
        )
        f_0 = alpha_x * (x - center_x) + alpha_y * (y - center_y)

        npt.assert_almost_equal(output, out_spp - f_0, decimal=8)

    def test_derivatives(self):
        tangential_stretch = 5
        radial_stretch = 1
        curvature = 1.0 / 10
        direction = 0.3
        center_x = 0
        center_y = 0
        x, y = 1, 1
        theta_E, gamma, center_x_spp, center_y_spp = self.model.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        f_x, f_y = self.spp.derivatives(
            x, y, theta_E, gamma, center_x_spp, center_y_spp
        )
        f_x0, f_y0 = self.spp.derivatives(
            center_x, center_y, theta_E, gamma, center_x_spp, center_y_spp
        )
        f_x_new, f_y_new = self.model.derivatives(
            x,
            y,
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
            center_x,
            center_y,
        )
        npt.assert_almost_equal(f_x_new, f_x - f_x0, decimal=8)
        npt.assert_almost_equal(f_y_new, f_y - f_y0, decimal=8)

    def test_hessian(self):
        lens = LensModel(lens_model_list=["CURVED_ARC_SPP"])
        center_x, center_y = 0, 0
        tangential_stretch = 10
        radial_stretch = 1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 2, 3
        tangential_stretch = 10
        radial_stretch = 1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = 3
        radial_stretch = -1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = -3
        radial_stretch = -1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = 10.4
        radial_stretch = 0.6
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "sibirrer"


import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Dipole", "DipoleUtil"]


class Dipole(LensProfileBase):
    """Class for dipole response of two massive bodies (experimental)"""

    param_names = ["com_x", "com_y", "phi_dipole", "coupling"]
    lower_limit_default = {
        "com_x": -100,
        "com_y": -100,
        "phi_dipole": -10,
        "coupling": -10,
    }
    upper_limit_default = {"com_x": 100, "com_y": 100, "phi_dipole": 10, "coupling": 10}

    def function(self, x, y, com_x, com_y, phi_dipole, coupling):
        # coordinate shift
        x_shift = x - com_x
        y_shift = y - com_y

        # rotation angle
        sin_phi = np.sin(phi_dipole)
        cos_phi = np.cos(phi_dipole)
        x_ = cos_phi * x_shift + sin_phi * y_shift
        # y_ = -sin_phi*x_shift + cos_phi*y_shift
        # r = np.sqrt(x_**2 + y_**2)

        # f_ = coupling**2 * (x_/y_)**2  # np.sqrt(np.abs(y_)/r) * np.abs(y_)
        # f_ = coupling * np.abs(x_)
        f_ = np.zeros_like(x_)
        return f_

    def derivatives(self, x, y, com_x, com_y, phi_dipole, coupling):
        # coordinate shift
        x_shift = x - com_x
        y_shift = y - com_y

        # rotation angle
        sin_phi = np.sin(phi_dipole)
        cos_phi = np.cos(phi_dipole)
        x_ = cos_phi * x_shift + sin_phi * y_shift
        y_ = -sin_phi * x_shift + cos_phi * y_shift

        f_x_prim = coupling * x_ / np.sqrt(x_**2 + y_**2)
        f_y_prim = np.zeros_like(x_)
        # rotate back
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, com_x, com_y, phi_dipole, coupling):
        # coordinate shift
        x_shift = x - com_x
        y_shift = y - com_y

        # rotation angle
        sin_phi = np.sin(phi_dipole)
        cos_phi = np.cos(phi_dipole)
        x_ = cos_phi * x_shift + sin_phi * y_shift
        y_ = -sin_phi * x_shift + cos_phi * y_shift

        r = np.sqrt(x_**2 + y_**2)
        f_xx_prim = coupling * y_**2 / r**3
        f_xy_prim = -coupling * x_ * y_ / r**3
        f_yy_prim = np.zeros_like(x_)

        kappa = 1.0 / 2 * (f_xx_prim + f_yy_prim)
        gamma1_value = 1.0 / 2 * (f_xx_prim - f_yy_prim)
        gamma2_value = f_xy_prim
        # rotate back
        gamma1 = (
            np.cos(2 * phi_dipole) * gamma1_value
            - np.sin(2 * phi_dipole) * gamma2_value
        )
        gamma2 = (
            +np.sin(2 * phi_dipole) * gamma1_value
            + np.cos(2 * phi_dipole) * gamma2_value
        )

        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy


class DipoleUtil(object):
    """Pre-calculation of dipole properties."""

    @staticmethod
    def com(center1_x, center1_y, center2_x, center2_y, Fm):
        """
        :return: center of mass
        """
        com_x = (Fm * center1_x + center2_x) / (Fm + 1.0)
        com_y = (Fm * center1_y + center2_y) / (Fm + 1.0)
        return com_x, com_y

    @staticmethod
    def mass_ratio(theta_E, theta_E_sub):
        """Computes mass ration of the two clumps with given Einstein radius and power
        law slope (clump1/sub-clump) :param theta_E:

        :param theta_E_sub:
        :return:
        """
        return (theta_E / theta_E_sub) ** 2

    @staticmethod
    def angle(center1_x, center1_y, center2_x, center2_y):
        """Compute the rotation angle of the dipole :return:"""
        phi_G = np.arctan2(center2_y - center1_y, center2_x - center1_x)
        return phi_G



"""Routines that manipulate convolution kernels."""
import numpy as np
import copy
from scipy import ndimage
import lenstronomy.Util.util as util
import lenstronomy.Util.image_util as image_util
from lenstronomy.LightModel.Profiles.gaussian import Gaussian
from lenstronomy.LightModel.Profiles.moffat import Moffat
import lenstronomy.Util.multi_gauss_expansion as mge
from lenstronomy.GalKin import velocity_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def de_shift_kernel(kernel, shift_x, shift_y, iterations=20, fractional_step_size=1):
    """De-shifts a shifted kernel to the center of a pixel. This is performed
    iteratively.

    The input kernel is the solution of a linear interpolated shift of a sharper kernel
    centered in the middle of the pixel. To find the de-shifted kernel, we perform an
    iterative correction of proposed de-shifted kernels and compare its shifted version
    with the input kernel.

    :param kernel: (shifted) kernel, e.g. a star in an image that is not centered in the
        pixel grid
    :param shift_x: x-offset relative to the center of the pixel (sub-pixel shift)
    :param shift_y: y-offset relative to the center of the pixel (sub-pixel shift)
    :param iterations: number of repeated iterations of shifting a new de-shifted kernel
        and apply corrections
    :param fractional_step_size: float (0, 1] correction factor relative to previous
        proposal (can be used for stability
    :return: de-shifted kernel such that the interpolated shift boy (shift_x, shift_y)
        results in the input kernel
    """
    nx, ny = np.shape(kernel)
    kernel_new = (
        np.zeros((nx + 2, ny + 2))
        + (kernel[0, 0] + kernel[0, -1] + kernel[-1, 0] + kernel[-1, -1]) / 4.0
    )
    kernel_new[1:-1, 1:-1] = kernel
    int_shift_x = int(round(shift_x))
    frac_x_shift = shift_x - int_shift_x
    int_shift_y = int(round(shift_y))
    frac_y_shift = shift_y - int_shift_y
    kernel_init = copy.deepcopy(kernel_new)
    kernel_init_shifted = copy.deepcopy(
        ndimage.shift(kernel_init, shift=[int_shift_y, int_shift_x], order=1)
    )
    kernel_new = ndimage.shift(kernel_new, shift=[int_shift_y, int_shift_x], order=1)
    norm = np.sum(kernel_init_shifted)
    for i in range(iterations):
        kernel_shifted_inv = ndimage.shift(
            kernel_new, shift=[-frac_y_shift, -frac_x_shift], order=1
        )
        delta = kernel_init_shifted - kernel_norm(kernel_shifted_inv) * norm
        kernel_new += delta * fractional_step_size
        kernel_new = kernel_norm(kernel_new) * norm
    return kernel_new[1:-1, 1:-1]


@export
def center_kernel(kernel, iterations=20):
    """Given a kernel that might not be perfectly centered, this routine computes its
    light weighted center and then moves the center in an iterative process such that it
    is centered.

    :param kernel: 2d array (odd numbers)
    :param iterations: int, number of iterations
    :return: centered kernel
    """
    kernel = kernel_norm(kernel)
    nx, ny = np.shape(kernel)
    if nx % 2 == 0:
        raise ValueError("kernel needs odd number of pixels")
    # make coordinate grid of kernel
    x_grid, y_grid = util.make_grid(nx, deltapix=1, left_lower=False)
    # compute 1st moments to get light weighted center
    x_w = np.sum(kernel * util.array2image(x_grid))
    y_w = np.sum(kernel * util.array2image(y_grid))
    # de-shift kernel
    kernel_centered = de_shift_kernel(
        kernel, shift_x=-x_w, shift_y=-y_w, iterations=iterations
    )
    return kernel_norm(kernel_centered)


@export
def kernel_norm(kernel):
    """

    :param kernel:
    :return: normalisation of the psf kernel
    """
    norm = np.sum(np.array(kernel))
    kernel /= norm
    return kernel


@export
def subgrid_kernel(kernel, subgrid_res, odd=False, num_iter=100):
    """Creates a higher resolution kernel with subgrid resolution as an interpolation of
    the original kernel in an iterative approach.

    :param kernel: initial kernel
    :type kernel: 2d numpy array with square odd size
    :param subgrid_res: subgrid resolution required
    :type subgrid_res: integer
    :param odd: forces odd axis size return (-1 in size if even)
    :type odd: boolean
    :param num_iter: number of iterations in the de-shifting and enhancement
    :type num_iter: integer
    :return: kernel with higher resolution (larger)
    :rtype: 2d numpy array with n x subgrid size (-1 if result is even and odd=True)
    """
    subgrid_res = int(subgrid_res)
    if subgrid_res == 1:
        return kernel
    nx, ny = np.shape(kernel)
    d_x = 1.0 / nx
    x_in = np.linspace(d_x / 2, 1 - d_x / 2, nx)
    d_y = 1.0 / nx
    y_in = np.linspace(d_y / 2, 1 - d_y / 2, ny)
    nx_new = nx * subgrid_res
    ny_new = ny * subgrid_res
    if odd is True:
        if nx_new % 2 == 0:
            nx_new -= 1
        if ny_new % 2 == 0:
            ny_new -= 1

    d_x_new = 1.0 / nx_new
    d_y_new = 1.0 / ny_new
    x_out = np.linspace(d_x_new / 2.0, 1 - d_x_new / 2.0, nx_new)
    y_out = np.linspace(d_y_new / 2.0, 1 - d_y_new / 2.0, ny_new)
    kernel_input = copy.deepcopy(kernel)
    kernel_subgrid = image_util.re_size_array(x_in, y_in, kernel_input, x_out, y_out)
    kernel_subgrid = kernel_norm(kernel_subgrid)
    for i in range(max(num_iter, 1)):
        # given a proposition, re-size it to original pixel size
        if subgrid_res % 2 == 0:
            kernel_pixel = averaging_even_kernel(kernel_subgrid, subgrid_res)
        else:
            kernel_pixel = util.averaging(kernel_subgrid, numGrid=nx_new, numPix=nx)
        delta = kernel - kernel_pixel
        temp_kernel = kernel_input + delta
        kernel_subgrid = image_util.re_size_array(
            x_in, y_in, temp_kernel, x_out, y_out
        )  # /norm_subgrid
        kernel_subgrid = kernel_norm(kernel_subgrid)
        kernel_input = temp_kernel

    # from scipy.ndimage import zoom
    # ratio = subgrid_res
    # kernel_subgrid = zoom(kernel, ratio, order=4) / ratio ** 2
    # print(np.shape(kernel_subgrid))
    # whatever has not been matched is added to zeroth order (in squares of the undersampled PSF)
    if subgrid_res % 2 == 0:
        return kernel_subgrid
    kernel_pixel = util.averaging(kernel_subgrid, numGrid=nx_new, numPix=nx)
    kernel_pixel = kernel_norm(kernel_pixel)
    delta_kernel = kernel_pixel - kernel_norm(kernel)
    id = np.ones((subgrid_res, subgrid_res))
    delta_kernel_sub = np.kron(delta_kernel, id) / subgrid_res**2
    return kernel_norm(kernel_subgrid - delta_kernel_sub)


@export
def kernel_pixelsize_change(kernel, deltaPix_in, deltaPix_out):
    """Change the pixel size of a given kernel.

    :param kernel:
    :param deltaPix_in:
    :param deltaPix_out:
    :return:
    """
    numPix = len(kernel)
    numPix_new = int(round(numPix * deltaPix_in / deltaPix_out))
    if numPix_new % 2 == 0:
        numPix_new -= 1
    x_in = np.linspace(
        -(numPix - 1) / 2 * deltaPix_in, (numPix - 1) / 2 * deltaPix_in, numPix
    )
    x_out = np.linspace(
        -(numPix_new - 1) / 2 * deltaPix_out,
        (numPix_new - 1) / 2 * deltaPix_out,
        numPix_new,
    )
    kernel_out = image_util.re_size_array(x_in, x_in, kernel, x_out, x_out)
    kernel_out = kernel_norm(kernel_out)
    return kernel_out


@export
def cut_psf(psf_data, psf_size, normalisation=True):
    """Cut the psf properly.

    :param psf_data: image of PSF
    :param psf_size: size of psf
    :return: re-sized and re-normalized PSF
    """
    kernel = image_util.cut_edges(psf_data, psf_size)
    if normalisation is True:
        kernel = kernel_norm(kernel)
    return kernel


@export
def pixel_kernel(point_source_kernel, subgrid_res=7):
    """Converts a pixelised kernel of a point source to a kernel representing a uniform
    extended pixel.

    :param point_source_kernel:
    :param subgrid_res:
    :return: convolution kernel for an extended pixel
    """
    kernel_subgrid = subgrid_kernel(point_source_kernel, subgrid_res, num_iter=10)
    kernel_size = len(point_source_kernel)
    kernel_pixel = np.zeros((kernel_size * subgrid_res, kernel_size * subgrid_res))
    for i in range(subgrid_res):
        k_x = int((kernel_size - 1) / 2 * subgrid_res + i)
        for j in range(subgrid_res):
            k_y = int((kernel_size - 1) / 2 * subgrid_res + j)
            kernel_pixel = image_util.add_layer2image(
                kernel_pixel, k_x, k_y, kernel_subgrid
            )
    kernel_pixel = util.averaging(
        kernel_pixel, numGrid=kernel_size * subgrid_res, numPix=kernel_size
    )
    return kernel_norm(kernel_pixel)


@export
def kernel_average_pixel(kernel_super, supersampling_factor):
    """Computes the effective convolution kernel assuming a uniform surface brightness
    on the scale of a pixel.

    :param kernel_super: supersampled PSF of a point source (odd number per axis
    :param supersampling_factor: supersampling factor (int)
    :return:
    """
    kernel_sum = np.sum(kernel_super)
    kernel_size = int(round(len(kernel_super) / float(supersampling_factor) + 0.5))
    if kernel_size % 2 == 0:
        kernel_size += 1
    n_high = kernel_size * supersampling_factor
    if n_high % 2 == 0:
        n_high += 1
    kernel_pixel = np.zeros((n_high, n_high))
    for i in range(supersampling_factor):
        k_x = int((kernel_size - 1) / 2 * supersampling_factor + i)
        for j in range(supersampling_factor):
            k_y = int((kernel_size - 1) / 2 * supersampling_factor + j)
            kernel_pixel = image_util.add_layer2image(
                kernel_pixel, k_x, k_y, kernel_super
            )

    if supersampling_factor % 2 == 0:
        kernel_pixel = averaging_even_kernel(kernel_pixel, supersampling_factor)
    else:
        kernel_pixel = util.averaging(kernel_pixel, numGrid=n_high, numPix=kernel_size)
    kernel_pixel /= np.sum(kernel_pixel)
    return kernel_pixel * kernel_sum


@export
def kernel_gaussian(num_pix, delta_pix, fwhm):
    """Gaussian kernel.

    :param num_pix: number of pixels
    :param delta_pix: pixel scale
    :param fwhm: full width at half maximum
    :return: 2d kernel
    """
    sigma = util.fwhm2sigma(fwhm)
    # if kernel_numPix % 2 == 0:
    #    kernel_numPix += 1
    x_grid, y_grid = util.make_grid(num_pix, delta_pix)
    gaussian = Gaussian()
    kernel = gaussian.function(
        x_grid, y_grid, amp=1.0, sigma=sigma, center_x=0, center_y=0
    )
    kernel /= np.sum(kernel)
    kernel = util.array2image(kernel)
    return kernel


def kernel_moffat(num_pix, delta_pix, fwhm, moffat_beta):
    """Moffat kernel.

    :param delta_pix: pixel scale of kernel
    :param num_pix: number of pixels per axis of the kernel
    :param fwhm: full width at half maximum
    :param moffat_beta: beta of Moffat profile
    :return: 2d kernel
    """
    alpha = velocity_util.moffat_fwhm_alpha(fwhm, moffat_beta)
    x, y = util.make_grid(numPix=num_pix, deltapix=delta_pix)
    moffat = Moffat()
    kernel = moffat.function(x=x, y=y, amp=1, alpha=alpha, beta=moffat_beta)
    kernel /= np.sum(kernel)
    kernel = util.array2image(kernel)
    return kernel


@export
def split_kernel(
    kernel_super, supersampling_kernel_size, supersampling_factor, normalized=True
):
    """Pixel kernel and subsampling kernel such that the convolution of both applied on
    an image can be performed, i.e. smaller subsampling PSF and hole in larger PSF.

    :param kernel_super: super-sampled kernel
    :param supersampling_kernel_size: size of super-sampled PSF in units of degraded
        pixels
    :param normalized: boolean, if True returns a split kernel that is area normalized=1
        representing a convolution kernel
    :return: degraded kernel with hole and super-sampled kernel
    """
    if supersampling_factor <= 1:
        raise ValueError(
            "To split a kernel, the supersampling_factor needs to be > 1, given %s"
            % supersampling_factor
        )
    if supersampling_kernel_size % 2 == 0:
        raise ValueError("supersampling_kernel_size needs to be an odd number!")
    n_super = len(kernel_super)
    n_sub = supersampling_kernel_size * supersampling_factor
    if n_sub % 2 == 0:
        n_sub += 1
    if n_sub > n_super:
        n_sub = n_super

    kernel_hole = copy.deepcopy(kernel_super)
    n_min = int((n_super - 1) / 2 - (n_sub - 1) / 2)
    n_max = int((n_super - 1) / 2 + (n_sub - 1) / 2 + 1)
    kernel_hole[n_min:n_max, n_min:n_max] = 0
    kernel_hole_resized = degrade_kernel(
        kernel_hole, degrading_factor=supersampling_factor
    )
    kernel_subgrid_cut = kernel_super[n_min:n_max, n_min:n_max]
    if normalized is True:
        flux_subsampled = np.sum(kernel_subgrid_cut)
        flux_hole = np.sum(kernel_hole_resized)
        if flux_hole > 0:
            kernel_hole_resized *= (1.0 - flux_subsampled) / np.sum(kernel_hole_resized)
        else:
            kernel_subgrid_cut /= np.sum(kernel_subgrid_cut)
    else:
        kernel_hole_resized /= supersampling_factor**2
    return kernel_hole_resized, kernel_subgrid_cut


@export
def degrade_kernel(kernel_super, degrading_factor):
    """

    :param kernel_super: higher resolution kernel (odd number per axis)
    :param degrading_factor: degrading factor (effectively the super-sampling resolution of the kernel given
    :return: degraded kernel with odd axis number with the sum of the flux/values in the kernel being preserved
    """
    if degrading_factor == 1:
        return kernel_super
    if degrading_factor % 2 == 0:
        kernel_low_res = averaging_even_kernel(kernel_super, degrading_factor)
    else:
        kernel_low_res = averaging_odd_kernel(kernel_super, degrading_factor)
    # degrading_factor**2  # multiplicative factor added when providing flux conservation
    kernel_low_res *= degrading_factor**2
    return kernel_low_res


def averaging_odd_kernel(kernel_super, degrading_factor):
    """"""
    n_kernel = len(kernel_super)
    numPix = int(round(n_kernel / degrading_factor + 0.5))
    if numPix % 2 == 0:
        numPix += 1
    n_high = numPix * degrading_factor

    kernel_super_ = np.zeros((n_high, n_high))
    i_start = int((n_high - n_kernel) / 2)
    kernel_super_[
        i_start : i_start + n_kernel, i_start : i_start + n_kernel
    ] = kernel_super
    kernel_low_res = util.averaging(kernel_super_, numGrid=n_high, numPix=numPix)
    return kernel_low_res


@export
def averaging_even_kernel(kernel_high_res, subgrid_res):
    """Makes a lower resolution kernel based on the kernel_high_res (odd numbers) and
    the subgrid_res (even number), both meant to be centered.

    :param kernel_high_res: high resolution kernel with even subsampling resolution,
        centered
    :param subgrid_res: subsampling resolution (even number)
    :return: averaged undersampling kernel
    """
    n_kernel_high_res = len(kernel_high_res)
    n_low = int(round(n_kernel_high_res / subgrid_res + 0.5))
    if n_low % 2 == 0:
        n_low += 1
    n_high = int(n_low * subgrid_res - 1)
    assert n_high % 2 == 1
    if n_high == n_kernel_high_res:
        kernel_high_res_edges = kernel_high_res
    else:
        i_start = int((n_high - n_kernel_high_res) / 2)
        kernel_high_res_edges = np.zeros((n_high, n_high))
        kernel_high_res_edges[i_start:-i_start, i_start:-i_start] = kernel_high_res
    kernel_low_res = np.zeros((n_low, n_low))
    # adding pixels that are fully within a single re-binned pixel
    for i in range(subgrid_res - 1):
        for j in range(subgrid_res - 1):
            kernel_low_res += kernel_high_res_edges[i::subgrid_res, j::subgrid_res]
    # adding half of a pixel that has over-lap with two pixels
    i = subgrid_res - 1
    for j in range(subgrid_res - 1):
        kernel_low_res[1:, :] += (
            kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
        )
        kernel_low_res[:-1, :] += (
            kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
        )
    j = subgrid_res - 1
    for i in range(subgrid_res - 1):
        kernel_low_res[:, 1:] += (
            kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
        )
        kernel_low_res[:, :-1] += (
            kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
        )
    # adding a quarter of a pixel value that is at the boarder of four pixels
    i = subgrid_res - 1
    j = subgrid_res - 1
    kernel_edge = kernel_high_res_edges[i::subgrid_res, j::subgrid_res]
    kernel_low_res[1:, 1:] += kernel_edge / 4
    kernel_low_res[:-1, 1:] += kernel_edge / 4
    kernel_low_res[1:, :-1] += kernel_edge / 4
    kernel_low_res[:-1, :-1] += kernel_edge / 4
    return kernel_low_res / subgrid_res**2


@export
def cutout_source(x_pos, y_pos, image, kernelsize, shift=True):
    """Cuts out point source (e.g. PSF estimate) out of image and shift it to the center
    of a pixel.

    :param x_pos:
    :param y_pos:
    :param image:
    :param kernelsize:
    :return:
    """
    if kernelsize % 2 == 0:
        raise ValueError("even pixel number kernel size not supported!")
    x_int = int(round(x_pos))
    y_int = int(round(y_pos))
    n = len(image)
    d = (kernelsize - 1) / 2
    x_max = int(np.minimum(x_int + d + 1, n))
    x_min = int(np.maximum(x_int - d, 0))
    y_max = int(np.minimum(y_int + d + 1, n))
    y_min = int(np.maximum(y_int - d, 0))
    image_cut = copy.deepcopy(image[y_min:y_max, x_min:x_max])
    shift_x = x_int - x_pos
    shift_y = y_int - y_pos
    if shift is True:
        kernel_shift = de_shift_kernel(image_cut, shift_x, shift_y, iterations=50)
    else:
        kernel_shift = image_cut
    kernel_final = np.zeros((kernelsize, kernelsize))

    k_l2_x = int((kernelsize - 1) / 2)
    k_l2_y = int((kernelsize - 1) / 2)

    xk_min = np.maximum(0, -x_int + k_l2_x)
    yk_min = np.maximum(0, -y_int + k_l2_y)
    xk_max = np.minimum(kernelsize, -x_int + k_l2_x + n)
    yk_max = np.minimum(kernelsize, -y_int + k_l2_y + n)

    kernel_final[yk_min:yk_max, xk_min:xk_max] = kernel_shift
    return kernel_final


@export
def fwhm_kernel(kernel):
    """

    :param kernel:
    :return:
    """
    n = len(kernel)
    center = (n - 1) / 2.0
    I_r = image_util.radial_profile(kernel, center=[center, center])
    if n % 2 == 0:
        raise ValueError("only works with odd number of pixels in kernel!")
    max_flux = kernel[int((n - 1) / 2), int((n - 1) / 2)]
    I_2 = max_flux / 2.0
    r = np.linspace(0, (n - 1) / 2, int((n + 1) / 2)) + 0.33
    for i in range(1, len(I_r)):
        if I_r[i] < I_2:
            fwhm_2 = (I_2 - I_r[i - 1]) / (I_r[i] - I_r[i - 1]) + r[i - 1]
            return fwhm_2 * 2
    raise ValueError(
        "The kernel did not drop to half the max value - fwhm not determined!"
    )


@export
def estimate_amp(data, x_pos, y_pos, psf_kernel):
    """Estimates the amplitude of a point source located at x_pos, y_pos.

    :param data:
    :param x_pos:
    :param y_pos:
    :param psf_kernel:
    :return:
    """
    numPix_x, numPix_y = np.shape(data)
    x_int = int(round(x_pos - 0.49999))
    y_int = int(round(y_pos - 0.49999))
    # TODO: make amplitude estimate not sucebtible to rounding effects on which pixels to chose to estimate the amplitude
    if x_int > 2 and x_int < numPix_x - 2 and y_int > 2 and y_int < numPix_y - 2:
        mean_image = max(np.sum(data[y_int - 2 : y_int + 3, x_int - 2 : x_int + 3]), 0)
        num = len(psf_kernel)
        center = int((num - 0.5) / 2)
        mean_kernel = np.sum(
            psf_kernel[center - 2 : center + 3, center - 2 : center + 3]
        )
        amp_estimated = mean_image / mean_kernel
    else:
        amp_estimated = 0
    return amp_estimated


@export
def mge_kernel(kernel, order=5):
    """Azimutal Multi-Gaussian expansion of a pixelized kernel.

    :param kernel: 2d numpy array
    :return:
    """
    # radial average
    n = len(kernel)
    center = (n - 1) / 2.0
    psf_r = image_util.radial_profile(kernel, center=[center, center])
    # MGE of radial average
    n_r = len(psf_r)
    r_array = np.linspace(start=0.0, stop=n_r - 1, num=n_r)
    amps, sigmas, norm = mge.mge_1d(r_array, psf_r, N=order, linspace=True)
    return amps, sigmas, norm


@export
def match_kernel_size(image, size):
    """Matching kernel/image to a dedicated size by either expanding the image with
    zeros at the edges or chopping of the edges.

    :param image: 2d array (square with odd number of pixels)
    :param size: integer (odd number)
    :return: image with matched size, either by cutting or by adding zeros in the
        outskirts
    """
    n = len(image)
    if n == size:
        return image
    image_copy = copy.deepcopy(image)
    if n > size:
        return image_copy[
            int((n - size) / 2) : int(n - (n - size) / 2),
            int((n - size) / 2) : int(n - (n - size) / 2),
        ]
    else:
        image_add = np.zeros((size, size))
        image_add[
            int((size - n) / 2) : int(size - (size - n) / 2),
            int((size - n) / 2) : int(size - (size - n) / 2),
        ] = image_copy
        return image_add



__author__ = "sibirrer"
from lenstronomy.LensModel.single_plane import SinglePlane
from lenstronomy.LensModel.LineOfSight.single_plane_los import SinglePlaneLOS
from lenstronomy.LensModel.MultiPlane.multi_plane import MultiPlane
from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from lenstronomy.Util import constants as const

__all__ = ["LensModel"]


class LensModel(object):
    """Class to handle an arbitrary list of lens models.

    This is the main lenstronomy LensModel API for all other modules.
    """

    def __init__(
        self,
        lens_model_list,
        z_lens=None,
        z_source=None,
        lens_redshift_list=None,
        cosmo=None,
        multi_plane=False,
        numerical_alpha_class=None,
        observed_convention_index=None,
        z_source_convention=None,
        cosmo_interp=False,
        z_interp_stop=None,
        num_z_interp=100,
        kwargs_interp=None,
        kwargs_synthesis=None,
    ):
        """

        :param lens_model_list: list of strings with lens model names
        :param z_lens: redshift of the deflector (only considered when operating in single plane mode).
         Is only needed for specific functions that require a cosmology.
        :param z_source: redshift of the source: Needed in multi_plane option only,
         not required for the core functionalities in the single plane mode.
        :param lens_redshift_list: list of deflector redshift (corresponding to the lens model list),
         only applicable in multi_plane mode.
        :param cosmo: instance of the astropy cosmology class. If not specified, uses the default cosmology.
        :param multi_plane: bool, if True, uses multi-plane mode. Default is False.
        :param numerical_alpha_class: an instance of a custom class for use in TabulatedDeflections() lens model
         (see documentation in Profiles/numerical_deflections)
        :param kwargs_interp: interpolation keyword arguments specifying the numerics.
         See description in the Interpolate() class. Only applicable for 'INTERPOL' and 'INTERPOL_SCALED' models.
        :param observed_convention_index: a list of indices, corresponding to the lens_model_list element with same
         index, where the 'center_x' and 'center_y' kwargs correspond to observed (lensed) positions, not physical
         positions. The code will compute the physical locations when performing computations
        :param z_source_convention: float, redshift of a source to define the reduced deflection angles of the lens
         models. If None, 'z_source' is used.
        :param cosmo_interp: boolean (only employed in multi-plane mode), interpolates astropy.cosmology distances for
         faster calls when accessing several lensing planes
        :param z_interp_stop: (only in multi-plane with cosmo_interp=True); maximum redshift for distance interpolation
         This number should be higher or equal the maximum of the source redshift and/or the z_source_convention
        :param num_z_interp: (only in multi-plane with cosmo_interp=True); number of redshift bins for interpolating
         distances
        """
        self.lens_model_list = lens_model_list
        self.z_lens = z_lens
        self.z_source = z_source
        self._z_source_convention = z_source_convention
        self.redshift_list = lens_redshift_list

        if cosmo is None:
            from astropy.cosmology import default_cosmology

            cosmo = default_cosmology.get()
        self.cosmo = cosmo

        # Are there line-of-sight corrections?
        permitted_los_models = ["LOS", "LOS_MINIMAL"]
        los_models = [
            (i, model)
            for (i, model) in enumerate(lens_model_list)
            if model in permitted_los_models
        ]
        if len(los_models) == 0:
            los_effects = False
        elif len(los_models) == 1:
            los_effects = True
            index_los, los_model = los_models[0]
        else:
            raise ValueError(
                "You can only have one model for line-of-sight corrections."
            )

        # Multi-plane or single-plane lensing?
        self.multi_plane = multi_plane
        if multi_plane is True:
            if z_source is None:
                raise ValueError(
                    "z_source needs to be set for multi-plane lens modelling."
                )
            if los_effects is True:
                raise ValueError(
                    "LOS effects and multi-plane lensing are incompatible."
                )
            self.lens_model = MultiPlane(
                z_source,
                lens_model_list,
                lens_redshift_list,
                cosmo=cosmo,
                numerical_alpha_class=numerical_alpha_class,
                observed_convention_index=observed_convention_index,
                z_source_convention=z_source_convention,
                cosmo_interp=cosmo_interp,
                z_interp_stop=z_interp_stop,
                num_z_interp=num_z_interp,
                kwargs_interp=kwargs_interp,
                kwargs_synthesis=kwargs_synthesis,
            )
        else:
            if los_effects is True:
                self.lens_model = SinglePlaneLOS(
                    lens_model_list,
                    index_los=index_los,
                    numerical_alpha_class=numerical_alpha_class,
                    lens_redshift_list=lens_redshift_list,
                    z_source_convention=z_source_convention,
                    kwargs_interp=kwargs_interp,
                    kwargs_synthesis=kwargs_synthesis,
                )
            else:
                self.lens_model = SinglePlane(
                    lens_model_list,
                    numerical_alpha_class=numerical_alpha_class,
                    lens_redshift_list=lens_redshift_list,
                    z_source_convention=z_source_convention,
                    kwargs_interp=kwargs_interp,
                    kwargs_synthesis=kwargs_synthesis,
                )

        if z_lens is not None and z_source is not None:
            self._lensCosmo = LensCosmo(z_lens, z_source, cosmo=cosmo)

    def ray_shooting(self, x, y, kwargs, k=None):
        """Maps image to source position (inverse deflection)

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: source plane positions corresponding to (x, y) in the image plane
        """
        return self.lens_model.ray_shooting(x, y, kwargs, k=k)

    def fermat_potential(
        self, x_image, y_image, kwargs_lens, x_source=None, y_source=None
    ):
        """Fermat potential (negative sign means earlier arrival time) for Multi-plane
        lensing, it computes the effective Fermat potential (derived from the arrival
        time and subtracted off the time-delay distance for the given cosmology). The
        units are given in arcsecond square.

        :param x_image: image position
        :param y_image: image position
        :param x_source: source position
        :param y_source: source position
        :param kwargs_lens: list of keyword arguments of lens model parameters matching
            the lens model classes
        :return: fermat potential in arcsec**2 without geometry term (second part of Eqn
            1 in Suyu et al. 2013) as a list
        """
        if hasattr(self.lens_model, "fermat_potential"):
            return self.lens_model.fermat_potential(
                x_image, y_image, kwargs_lens, x_source, y_source
            )
        elif hasattr(self.lens_model, "arrival_time") and hasattr(self, "_lensCosmo"):
            dt = self.lens_model.arrival_time(x_image, y_image, kwargs_lens)
            fermat_pot_eff = (
                dt
                * const.c
                / self._lensCosmo.ddt
                / const.Mpc
                * const.day_s
                / const.arcsec**2
            )
            return fermat_pot_eff
        else:
            raise ValueError(
                "In multi-plane lensing you need to provide a specific z_lens and z_source for which the "
                "effective Fermat potential is evaluated"
            )

    def arrival_time(
        self, x_image, y_image, kwargs_lens, kappa_ext=0, x_source=None, y_source=None
    ):
        """Arrival time of images relative to a straight line without lensing. Negative
        values correspond to images arriving earlier, and positive signs correspond to
        images arriving later.

        :param x_image: image position
        :param y_image: image position
        :param kwargs_lens: lens model parameter keyword argument list
        :param kappa_ext: external convergence contribution not accounted in the lens
            model that leads to the same observables in position and relative fluxes but
            rescales the time delays
        :param x_source: source position (optional), otherwise computed with ray-tracing
        :param y_source: source position (optional), otherwise computed with ray-tracing
        :return: arrival time of image positions in units of days
        """
        if hasattr(self.lens_model, "arrival_time"):
            arrival_time = self.lens_model.arrival_time(x_image, y_image, kwargs_lens)
        else:
            fermat_pot = self.lens_model.fermat_potential(
                x_image, y_image, kwargs_lens, x_source=x_source, y_source=y_source
            )
            if not hasattr(self, "_lensCosmo"):
                raise ValueError(
                    "LensModel class was not initialized with lens and source redshifts!"
                )
            arrival_time = self._lensCosmo.time_delay_units(fermat_pot)
        arrival_time *= 1 - kappa_ext
        return arrival_time

    def potential(self, x, y, kwargs, k=None):
        """Lensing potential.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing potential in units of arcsec^2
        """
        return self.lens_model.potential(x, y, kwargs, k=k)

    def alpha(self, x, y, kwargs, k=None, diff=None):
        """Deflection angles.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :param diff: None or float. If set, computes the deflection as a finite
            numerical differential of the lensing potential. This differential is only
            applicable in the single lensing plane where the form of the lensing
            potential is analytically known
        :return: deflection angles in units of arcsec
        """
        if diff is None:
            return self.lens_model.alpha(x, y, kwargs, k=k)
        elif self.multi_plane is False:
            return self._deflection_differential(x, y, kwargs, k=k, diff=diff)
        else:
            raise ValueError(
                "numerical differentiation of lensing potential is not available in the multi-plane "
                "setting as analytical form of lensing potential is not available."
            )

    def hessian(self, x, y, kwargs, k=None, diff=None, diff_method="square"):
        """Hessian matrix.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :param diff: float, scale over which the finite numerical differential is
            computed. If None, then using the exact (if available) differentials.
        :param diff_method: string, 'square' or 'cross', indicating whether finite
            differentials are computed from a cross or a square of points around (x, y)
        :return: f_xx, f_xy, f_yx, f_yy components
        """
        if diff is None:
            return self.lens_model.hessian(x, y, kwargs, k=k)
        elif diff_method == "square":
            return self._hessian_differential_square(x, y, kwargs, k=k, diff=diff)
        elif diff_method == "cross":
            return self._hessian_differential_cross(x, y, kwargs, k=k, diff=diff)
        else:
            raise ValueError(
                'diff_method %s not supported. Chose among "square" or "cross".'
                % diff_method
            )

    def kappa(self, x, y, kwargs, k=None, diff=None, diff_method="square"):
        """Lensing convergence k = 1/2 laplacian(phi)

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :param diff: float, scale over which the finite numerical differential is
            computed. If None, then using the exact (if available) differentials.
        :param diff_method: string, 'square' or 'cross', indicating whether finite
            differentials are computed from a cross or a square of points around (x, y)
        :return: lensing convergence
        """

        f_xx, f_xy, f_yx, f_yy = self.hessian(
            x, y, kwargs, k=k, diff=diff, diff_method=diff_method
        )
        kappa = 1.0 / 2 * (f_xx + f_yy)
        return kappa

    def curl(self, x, y, kwargs, k=None, diff=None, diff_method="square"):
        """
        curl computation F_xy - F_yx

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :param diff: float, scale over which the finite numerical differential is computed. If None, then using the
         exact (if available) differentials.
        :param diff_method: string, 'square' or 'cross', indicating whether finite differentials are computed from a
         cross or a square of points around (x, y)
        :return: curl at position (x, y)
        """
        f_xx, f_xy, f_yx, f_yy = self.hessian(
            x, y, kwargs, k=k, diff=diff, diff_method=diff_method
        )
        return f_xy - f_yx

    def gamma(self, x, y, kwargs, k=None, diff=None, diff_method="square"):
        """
        shear computation
        g1 = 1/2(d^2phi/dx^2 - d^2phi/dy^2)
        g2 = d^2phi/dxdy

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :param diff: float, scale over which the finite numerical differential is computed. If None, then using the
         exact (if available) differentials.
        :param diff_method: string, 'square' or 'cross', indicating whether finite differentials are computed from a
         cross or a square of points around (x, y)
        :return: gamma1, gamma2
        """

        f_xx, f_xy, f_yx, f_yy = self.hessian(
            x, y, kwargs, k=k, diff=diff, diff_method=diff_method
        )
        gamma1 = 1.0 / 2 * (f_xx - f_yy)
        gamma2 = f_xy
        return gamma1, gamma2

    def magnification(self, x, y, kwargs, k=None, diff=None, diff_method="square"):
        """magnification.

        mag = 1/det(A)
        A = 1 - d^2phi/d_ij

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :param diff: float, scale over which the finite numerical differential is computed. If None, then using the
         exact (if available) differentials.
        :param diff_method: string, 'square' or 'cross', indicating whether finite differentials are computed from a
         cross or a square of points around (x, y)
        :return: magnification
        """

        f_xx, f_xy, f_yx, f_yy = self.hessian(
            x, y, kwargs, k=k, diff=diff, diff_method=diff_method
        )
        det_A = (1 - f_xx) * (1 - f_yy) - f_xy * f_yx
        return 1.0 / det_A  # attention, if dividing by zero

    def flexion(self, x, y, kwargs, k=None, diff=0.000001, hessian_diff=True):
        """Third derivatives (flexion)

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: int or None, if set, only evaluates the differential from one model
            component
        :param diff: numerical differential length of Flexion
        :param hessian_diff: boolean, if true also computes the numerical differential
            length of Hessian (optional)
        :return: f_xxx, f_xxy, f_xyy, f_yyy
        """
        if hessian_diff is not True:
            hessian_diff = None
        f_xx_dx, f_xy_dx, f_yx_dx, f_yy_dx = self.hessian(
            x + diff / 2, y, kwargs, k=k, diff=hessian_diff
        )
        f_xx_dy, f_xy_dy, f_yx_dy, f_yy_dy = self.hessian(
            x, y + diff / 2, kwargs, k=k, diff=hessian_diff
        )

        f_xx_dx_, f_xy_dx_, f_yx_dx_, f_yy_dx_ = self.hessian(
            x - diff / 2, y, kwargs, k=k, diff=hessian_diff
        )
        f_xx_dy_, f_xy_dy_, f_yx_dy_, f_yy_dy_ = self.hessian(
            x, y - diff / 2, kwargs, k=k, diff=hessian_diff
        )

        f_xxx = (f_xx_dx - f_xx_dx_) / diff
        f_xxy = (f_xx_dy - f_xx_dy_) / diff
        f_xyy = (f_xy_dy - f_xy_dy_) / diff
        f_yyy = (f_yy_dy - f_yy_dy_) / diff
        return f_xxx, f_xxy, f_xyy, f_yyy

    def set_static(self, kwargs):
        """Set this instance to a static lens model. This can improve the speed in
        evaluating lensing quantities at different positions but must not be used with
        different lens model parameters!

        :param kwargs: lens model keyword argument list
        :return: kwargs_updated (in case of image position convention in multiplane
            lensing this is changed)
        """
        return self.lens_model.set_static(kwargs)

    def set_dynamic(self):
        """Deletes cache for static setting and makes sure the observed convention in
        the position of lensing profiles in the multi-plane setting is enabled. Dynamic
        is the default setting of this class enabling an accurate computation of lensing
        quantities with different parameters in the lensing profiles.

        :return: None
        """
        self.lens_model.set_dynamic()

    def _deflection_differential(self, x, y, kwargs, k=None, diff=0.00001):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param kwargs: keyword argument list
        :param k: int or None, if set, only evaluates the differential from one model component
        :param diff: finite differential length
        :return: f_x, f_y
        """
        phi_dx = self.lens_model.potential(x + diff / 2, y, kwargs=kwargs, k=k)
        phi_dy = self.lens_model.potential(x, y + diff / 2, kwargs=kwargs, k=k)
        phi_dx_ = self.lens_model.potential(x - diff / 2, y, kwargs=kwargs, k=k)
        phi_dy_ = self.lens_model.potential(x, y - diff / 2, kwargs=kwargs, k=k)
        f_x = (phi_dx - phi_dx_) / diff
        f_y = (phi_dy - phi_dy_) / diff
        return f_x, f_y

    def _hessian_differential_cross(self, x, y, kwargs, k=None, diff=0.00001):
        """Computes the numerical differentials over a finite range for f_xx, f_yy, f_xy
        from f_x and f_y The differentials are computed along the cross centered at (x,
        y).

        :param x: x-coordinate
        :param y: y-coordinate
        :param kwargs: lens model keyword argument list
        :param k: int, list of bools or None, indicating a subset of lens models to be
            evaluated
        :param diff: float, scale of the finite differential (diff/2 in each direction
            used to compute the differential
        :return: f_xx, f_xy, f_yx, f_yy
        """
        alpha_ra_dx, alpha_dec_dx = self.alpha(x + diff / 2, y, kwargs, k=k)
        alpha_ra_dy, alpha_dec_dy = self.alpha(x, y + diff / 2, kwargs, k=k)

        alpha_ra_dx_, alpha_dec_dx_ = self.alpha(x - diff / 2, y, kwargs, k=k)
        alpha_ra_dy_, alpha_dec_dy_ = self.alpha(x, y - diff / 2, kwargs, k=k)

        dalpha_rara = (alpha_ra_dx - alpha_ra_dx_) / diff
        dalpha_radec = (alpha_ra_dy - alpha_ra_dy_) / diff
        dalpha_decra = (alpha_dec_dx - alpha_dec_dx_) / diff
        dalpha_decdec = (alpha_dec_dy - alpha_dec_dy_) / diff

        f_xx = dalpha_rara
        f_yy = dalpha_decdec
        f_xy = dalpha_radec
        f_yx = dalpha_decra
        return f_xx, f_xy, f_yx, f_yy

    def _hessian_differential_square(self, x, y, kwargs, k=None, diff=0.00001):
        """Computes the numerical differentials over a finite range for f_xx, f_yy, f_xy
        from f_x and f_y The differentials are computed on the square around (x, y).
        This minimizes curl.

        :param x: x-coordinate
        :param y: y-coordinate
        :param kwargs: lens model keyword argument list
        :param k: int, list of booleans or None, indicating a subset of lens models to
            be evaluated
        :param diff: float, scale of the finite differential (diff/2 in each direction
            used to compute the differential
        :return: f_xx, f_xy, f_yx, f_yy
        """
        alpha_ra_pp, alpha_dec_pp = self.alpha(x + diff / 2, y + diff / 2, kwargs, k=k)
        alpha_ra_pn, alpha_dec_pn = self.alpha(x + diff / 2, y - diff / 2, kwargs, k=k)

        alpha_ra_np, alpha_dec_np = self.alpha(x - diff / 2, y + diff / 2, kwargs, k=k)
        alpha_ra_nn, alpha_dec_nn = self.alpha(x - diff / 2, y - diff / 2, kwargs, k=k)

        f_xx = (alpha_ra_pp - alpha_ra_np + alpha_ra_pn - alpha_ra_nn) / diff / 2
        f_xy = (alpha_ra_pp - alpha_ra_pn + alpha_ra_np - alpha_ra_nn) / diff / 2
        f_yx = (alpha_dec_pp - alpha_dec_np + alpha_dec_pn - alpha_dec_nn) / diff / 2
        f_yy = (alpha_dec_pp - alpha_dec_pn + alpha_dec_np - alpha_dec_nn) / diff / 2

        return f_xx, f_xy, f_yx, f_yy

    def hessian_z1z2(self, z1, z2, theta_x, theta_y, kwargs_lens, diff=0.00000001):
        """Computes Hessian matrix when Observed at z1 with rays going to z2 with z1 <
        z2 for multi_plane.

        :param z1: Observer redshift
        :param z2: source redshift
        :param theta_x: angular position and direction of the ray
        :param theta_y: angular position and direction of the ray
        :param kwargs_lens: list of keyword arguments of lens model parameters matching
            the lens model classes
        :param diff: numerical differential step (float)
        :return: f_xx, f_xy, f_yx, f_yy
        """
        if self.multi_plane is False:
            raise ValueError("Hessian z1z2 need to be compute in multi-plane mode")
        if z1 >= z2:
            raise ValueError("z1 needs to be smaller than z2")

        f_xx, f_xy, f_yx, f_yy = self.lens_model.hessian_z1z2(
            z1, z2, theta_x, theta_y, kwargs_lens, diff=diff
        )

        return f_xx, f_xy, f_yx, f_yy



__author__ = "sibirrer"

import numpy as np
import lenstronomy.Util.util as util
from lenstronomy.LensModel.Profiles.sersic_utils import SersicUtil
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Sersic"]


class Sersic(SersicUtil, LensProfileBase):
    """
    this class contains functions to evaluate a Sersic mass profile: https://arxiv.org/pdf/astro-ph/0311559.pdf

    .. math::
        \\kappa(R) = \\kappa_{\\rm eff} \\exp \\left[ -b_n (R/R_{\\rm Sersic})^{\\frac{1}{n}}\\right]

    with :math:`b_{n}\\approx 1.999n-0.327`

    Examples
    --------

    Example for converting physical mass units into convergence units used in the definition of this profile.

    We first define an AstroPy cosmology instance and a LensCosmo class instance with a lens and source redshift.

    >>> from lenstronomy.Cosmo.lens_cosmo import LensCosmo
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.05)
    >>> lens_cosmo = LensCosmo(z_lens=0.5, z_source=1.5, cosmo=cosmo)

    We define the half-light radius R_sersic (arc seconds on the sky) and Sersic index n_sersic

    >>> R_sersic = 2
    >>> n_sersic = 4

    Here we compute k_eff, the convergence at the half-light radius R_sersic for a stellar mass in Msun

    >>> k_eff = lens_cosmo.sersic_m_star2k_eff(m_star=10**11.5, R_sersic=R_sersic, n_sersic=n_sersic)

    And here we perform the inverse calculation given k_eff to return the physical stellar mass.

    >>> m_star = lens_cosmo.sersic_k_eff2m_star(k_eff=k_eff, R_sersic=R_sersic, n_sersic=n_sersic)

    The lens model calculation uses angular units as arguments! So to execute a deflection angle calculation one uses

    >>> from lenstronomy.LensModel.Profiles.sersic import Sersic
    >>> sersic = Sersic()
    >>> alpha_x, alpha_y = sersic.derivatives(x=1, y=1, k_eff=k_eff, R_sersic=R_sersic, center_x=0, center_y=0)

    """

    param_names = ["k_eff", "R_sersic", "n_sersic", "center_x", "center_y"]
    lower_limit_default = {
        "k_eff": 0,
        "R_sersic": 0,
        "n_sersic": 0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "k_eff": 10,
        "R_sersic": 100,
        "n_sersic": 8,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, n_sersic, R_sersic, k_eff, center_x=0, center_y=0):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param n_sersic: Sersic index
        :param R_sersic: half light radius
        :param k_eff: convergence at half light radius
        :param center_x: x-center
        :param center_y: y-center
        :return:
        """

        n = n_sersic
        x_red = self._x_reduced(x, y, n, R_sersic, center_x, center_y)
        b = self.b_n(n)
        # hyper2f2_b = util.hyper2F2_array(2*n, 2*n, 1+2*n, 1+2*n, -b)
        hyper2f2_bx = util.hyper2F2_array(
            2 * n, 2 * n, 1 + 2 * n, 1 + 2 * n, -b * x_red
        )
        f_eff = np.exp(b) * R_sersic**2 / 2.0 * k_eff  # * hyper2f2_b
        f_ = f_eff * x_red ** (2 * n) * hyper2f2_bx  # / hyper2f2_b
        return f_

    def derivatives(self, x, y, n_sersic, R_sersic, k_eff, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function."""
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        if isinstance(r, int) or isinstance(r, float):
            r = max(self._s, r)
        else:
            r[r < self._s] = self._s
        alpha = -self.alpha_abs(x, y, n_sersic, R_sersic, k_eff, center_x, center_y)
        f_x = alpha * x_ / r
        f_y = alpha * y_ / r
        return f_x, f_y

    def hessian(self, x, y, n_sersic, R_sersic, k_eff, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        if isinstance(r, int) or isinstance(r, float):
            r = max(self._s, r)
        else:
            r[r < self._s] = self._s
        d_alpha_dr = self.d_alpha_dr(
            x, y, n_sersic, R_sersic, k_eff, center_x, center_y
        )
        alpha = -self.alpha_abs(x, y, n_sersic, R_sersic, k_eff, center_x, center_y)

        f_xx = -(d_alpha_dr / r + alpha / r**2) * x_**2 / r + alpha / r
        f_yy = -(d_alpha_dr / r + alpha / r**2) * y_**2 / r + alpha / r
        f_xy = -(d_alpha_dr / r + alpha / r**2) * x_ * y_ / r

        return f_xx, f_xy, f_xy, f_yy



__author__ = "lucateo"

# this file contains a class to compute the Ultra Light Dark Matter soliton profile
import numpy as np
from scipy.special import gamma, hyp2f1
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Uldm"]


class Uldm(LensProfileBase):
    """
    This class contains functions concerning the ULDM soliton density profile,
    whose good approximation is (see for example https://arxiv.org/pdf/1406.6586.pdf )

    .. math::

        \\rho = \\rho_0 (1 + a(\\theta/\\theta_c)^2)^{-\\beta}

    where :math:`\\theta_c` is the core radius, corresponding to the radius where the
    density drops by half its central value, :math: `\\beta` is the slope (called just slope
    in the parameters of this model), :math: `\\rho_0 = \\kappa_0 \\Sigma_c/D_lens`,
    and :math: `a` is a parameter, dependent on :math: `\\beta`, chosen such
    that :math: `\\theta_c` indeed corresponds to the radius where the density drops by half
    (simple math gives :math: `a = 0.5^{-1/\\beta} - 1` ).
    For an ULDM soliton profile without contributions to background potential, it
    turns out that :math: `\\beta = 8, a = 0.091`. We allow :math: `\\beta` to be
    different from 8 to model solitons which feel the influence of background
    potential (see 2105.10873)
    The profile has, as parameters:

    - kappa_0: central convergence
    - theta_c: core radius (in arcseconds)
    - slope: exponent entering the profile, default value is 8
    """

    _s = 0.000001  # numerical limit for minimal radius
    param_names = ["kappa_0", "theta_c", "slope", "center_x", "center_y"]
    lower_limit_default = {
        "kappa_0": 0,
        "theta_c": 0,
        "slope": 3.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "kappa_0": 1.0,
        "theta_c": 100,
        "slope": 10,
        "center_x": 100,
        "center_y": 100,
    }

    @staticmethod
    def rhotilde(kappa_0, theta_c, slope=8):
        """Computes the central density in angular units.

        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: central density in 1/arcsec
        """
        a_factor_sqrt = np.sqrt(0.5 ** (-1 / slope) - 1)
        num_factor = (
            gamma(slope) / gamma(slope - 1 / 2) * a_factor_sqrt / np.sqrt(np.pi)
        )
        return kappa_0 * num_factor / theta_c

    def function(self, x, y, kappa_0, theta_c, center_x=0, center_y=0, slope=8):
        """
        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: lensing potential (in arcsec^2)
        """
        from mpmath import hyp3f2

        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        a_factor_sqrt = np.sqrt((0.5) ** (-1.0 / slope) - 1)
        if np.isscalar(r) == True:
            hypgeom = float(
                kappa_0
                / 2
                * r**2
                * hyp3f2(1, 1, slope - 0.5, 2, 2, -((a_factor_sqrt * r / theta_c) ** 2))
            )
        else:
            hypgeom = np.array(
                [
                    kappa_0
                    / 2.0
                    * r_i**2.0
                    * hyp3f2(
                        1,
                        1,
                        slope - 0.5,
                        2,
                        2,
                        -((a_factor_sqrt * r_i / theta_c) ** 2.0),
                    )
                    for r_i in r
                ],
                dtype=float,
            )
        return hypgeom

    @staticmethod
    def alpha_radial(r, kappa_0, theta_c, slope=8):
        """Returns the radial part of the deflection angle.

        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :param r: radius where the deflection angle is computed
        :return: radial deflection angle
        """
        a_factor = 0.5 ** (-1.0 / slope) - 1
        prefactor = 2.0 / (2 * slope - 3) * kappa_0 * theta_c**2 / a_factor
        denominator_factor = (1 + a_factor * r**2 / theta_c**2) ** (slope - 3.0 / 2)
        return prefactor / r * (1 - 1 / denominator_factor)

    def derivatives(self, x, y, kappa_0, theta_c, center_x=0, center_y=0, slope=8):
        """Returns df/dx and df/dy of the function (lensing potential), which are the
        deflection angles.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection angle in x, deflection angle in y
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, 0.00000001)
        f_x = self.alpha_radial(R, kappa_0, theta_c, slope) * x_ / R
        f_y = self.alpha_radial(R, kappa_0, theta_c, slope) * y_ / R
        return f_x, f_y

    def hessian(self, x, y, kappa_0, theta_c, center_x=0, center_y=0, slope=8):
        """
        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, 0.00000001)
        a_factor = 0.5 ** (-1.0 / slope) - 1
        prefactor = 2.0 / (2 * slope - 3) * kappa_0 * theta_c**2 / a_factor
        # denominator factor
        denominator = 1 + a_factor * R**2 / theta_c**2
        factor1 = (
            (2 * slope - 3)
            * a_factor
            * denominator ** (1.0 / 2 - slope)
            / (theta_c**2 * R**2)
        )
        factor2 = 1 / R**4 * (1 - denominator ** (3.0 / 2 - slope))
        f_xx = prefactor * (factor1 * x_**2 + factor2 * (y_**2 - x_**2))
        f_yy = prefactor * (factor1 * y_**2 + factor2 * (x_**2 - y_**2))
        f_xy = prefactor * (factor1 * x_ * y_ - factor2 * 2 * x_ * y_)
        return f_xx, f_xy, f_xy, f_yy

    def density(self, R, kappa_0, theta_c, slope=8):
        """Three dimensional ULDM profile in angular units (rho0_physical = rho0_angular
        Sigma_crit / D_lens)

        :param R: radius of interest
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: rho(R) density in angular units
        """
        rhotilde = self.rhotilde(kappa_0, theta_c, slope)
        a_factor = 0.5 ** (-1.0 / slope) - 1
        return rhotilde / (1 + a_factor * (R / theta_c) ** 2) ** slope

    def density_lens(self, r, kappa_0, theta_c, slope=8):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: 3d radius
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: density rho(r)
        """
        return self.density(r, kappa_0, theta_c, slope)

    @staticmethod
    def kappa_r(R, kappa_0, theta_c, slope=8):
        """Convergence of the cored density profile. This routine is also for testing.

        :param R: radius (angular scale)
        :param kappa_0: convergence in the core
        :param theta_c: core radius
        :param slope: exponent entering the profile
        :return: convergence at r
        """
        a_factor = (0.5) ** (-1.0 / slope) - 1
        return kappa_0 * (1 + a_factor * (R / theta_c) ** 2) ** (1.0 / 2 - slope)

    def density_2d(self, x, y, kappa_0, theta_c, center_x=0, center_y=0, slope=8):
        """Projected two dimensional ULDM profile (convergence * Sigma_crit), but given
        our units convention for rho0, it is basically the convergence.

        :param x: x-coordinate
        :param y: y-coordinate
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: Epsilon(R) projected density at radius R
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        return self.kappa_r(R, kappa_0, theta_c, slope)

    def _mass_integral(self, x, slope=8):
        """Returns the analytic result of the integral appearing in mass expression.

        :param slope: exponent entering the profile
        :return: integral result
        """
        hypF = np.real(hyp2f1(3.0 / 2, slope, 5.0 / 2, -(x**2)))
        return 1.0 / 3 * x**3 * hypF

    def mass_3d(self, R, kappa_0, theta_c, slope=8):
        """Mass enclosed a 3d sphere or radius r.

        :param R: radius in arcseconds
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: mass of soliton in angular units
        """
        rhotilde = self.rhotilde(kappa_0, theta_c, slope)
        a_factor = 0.5 ** (-1.0 / slope) - 1
        prefactor = 4.0 * np.pi * rhotilde * theta_c**3 / (a_factor) ** (1.5)
        m_3d = prefactor * (
            self._mass_integral(R / theta_c * np.sqrt(a_factor), slope)
            - self._mass_integral(0, slope)
        )
        return m_3d

    def mass_3d_lens(self, r, kappa_0, theta_c, slope=8):
        """Mass enclosed a 3d sphere or radius r.

        :param r: radius over which the mass is computed
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: mass enclosed in 3D ball
        """
        m_3d = self.mass_3d(r, kappa_0, theta_c, slope)
        return m_3d

    def mass_2d(self, R, kappa_0, theta_c, slope=8):
        """Mass enclosed a 2d sphere or radius r.

        :param R: radius over which the mass is computed
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param slope: exponent entering the profile
        :return: mass enclosed in 2d sphere
        """
        m_2d = np.pi * R * self.alpha_radial(R, kappa_0, theta_c, slope)
        return m_2d
